/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/EnhPi3/Control.cpp,v $
 * $Date: 2002/05/29 10:04:21 $
 *
 Description:
	Functions to control execution of Pi3Web server on windows
	platforms.

	Most functions here return 1 for true/success, 0 for false and -1
	for errors. This is true unless otherwise stated.

\*____________________________________________________________________________*/
/* $SourceTop:$ */

#include <assert.h>
#include <windows.h>
#include <stdio.h>
#include <regstr.h>

#include "MiscUtil.h"
#include "SwapStr.h"

#include "Control.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
SC_HANDLE Internal_openServiceManager()
{
	SC_HANDLE schSCManager = OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	return schSCManager;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void Internal_closeServiceManager( SC_HANDLE schSCManager )
{
	CloseServiceHandle( schSCManager );
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return code: 1=service started, 0=service not running, -1=error
\*____________________________________________________________________________*/
int Internal_trackServiceState( HANDLE schService )
{
	SERVICE_STATUS ssStatus;

	/*
	** Get service status 
	*/
	if ( !QueryServiceStatus( schService, &ssStatus ) )
		{
		DWORD dwError = ::GetLastError();
		if ( dwError==ERROR_SERVICE_NOT_ACTIVE )
			return 0;
		else
			return -1;
		};
					
	/*
	** Check if service is clearly running or not running
	*/
	switch( ssStatus.dwCurrentState )
		{
		case SERVICE_RUNNING:	return 1;
		case SERVICE_STOPPED:	return 0;
		default:;
		};

	/*
	** Track service as it tries to start or stop
	*/
	DWORD dwOldCheckPoint = ssStatus.dwCheckPoint;
	for(;;)
		{
		/*
		** Service is running
		*/
		Sleep( ssStatus.dwWaitHint );
			
		if ( !QueryServiceStatus( schService, &ssStatus ) )
			{
			DWORD dwError = ::GetLastError();
			if ( dwError==ERROR_SERVICE_NOT_ACTIVE )
				return 0;
			else
				return -1;		/* error - unknown */
			};

		/*
		** Check if service is clearly running or not running
		*/
		switch( ssStatus.dwCurrentState )
			{
			case SERVICE_RUNNING:	return 1;
			case SERVICE_STOPPED:	return 0;
			default:;
			};

		/*
		** Check if service has stalled
		*/
		if ( dwOldCheckPoint >= ssStatus.dwCheckPoint )
			{ return -1; };		/* error - stalled */

		dwOldCheckPoint = ssStatus.dwCheckPoint;
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_isServiceInstalled( const char *pServiceName )
{
	assert( pServiceName );

	SC_HANDLE scH = Internal_openServiceManager();
	if ( !scH )
		{ return -1; };

	HANDLE schService = OpenService(
		scH, pServiceName, SERVICE_QUERY_STATUS );

	int iRet = 1;
	if ( schService )
		{
		CloseServiceHandle( schService );
		}
	else
		{
		iRet = GetLastError()==ERROR_SERVICE_DOES_NOT_EXIST ? 0 : -1;
		}

	Internal_closeServiceManager( scH );
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_installService( const char *pServiceName, const char *pPath )
{
	assert( pServiceName && pPath );

	SC_HANDLE scH = Internal_openServiceManager();
	if ( !scH )
		{ return -1; };

	DWORD dwStartType = SERVICE_AUTO_START;

	HANDLE schService = CreateService(
		scH,
		pServiceName,
		pServiceName,
		SERVICE_ALL_ACCESS,
		SERVICE_WIN32_OWN_PROCESS,
		dwStartType,
		SERVICE_ERROR_NORMAL,
		pPath,
		NULL,
		NULL,
		NULL,
		NULL,
		NULL );

	Internal_closeServiceManager( scH );

	if ( !schService )
		{ return -1; };

	CloseServiceHandle( schService );

	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_unInstallService( const char *pServiceName )
{
	assert( pServiceName );

	SC_HANDLE scH = Internal_openServiceManager();
	if ( !scH )
		{ return -1; };

	HANDLE schService = OpenService(
		scH, pServiceName, DELETE );

	Internal_closeServiceManager( scH );

	if ( !schService )
		{ return -1; };

	int iRet = DeleteService( schService );

	CloseServiceHandle( schService );

	return iRet ? 1 : -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_isServiceRunning( const char *pServiceName )
{
	assert( pServiceName );

	SC_HANDLE scH = Internal_openServiceManager();
	if ( !scH )
		{ return -1; };

	HANDLE schService = OpenService(
		scH, pServiceName, SERVICE_QUERY_STATUS );

	Internal_closeServiceManager( scH );
			
	if ( !schService )
		{ return -1; };

	int iRet = Internal_trackServiceState( schService );

	CloseServiceHandle( schService );

	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_startService( const char *pServiceName )
{
	assert( pServiceName );

	SC_HANDLE scH = Internal_openServiceManager();
	if ( !scH )
		{ return -1; };

	HANDLE schService = ::OpenService(
		scH, pServiceName, SERVICE_START | SERVICE_QUERY_STATUS );

	Internal_closeServiceManager( scH );
			
	if ( !schService )
		{ return -1; };

	/*
	** Try to start service
	*/
	if ( !::StartService( schService, 0, NULL ) )
		{ goto start_service_fail; };

	/*
	** Track service while it starts
	*/
	if ( Internal_trackServiceState( schService )!=1 )
		{ goto start_service_fail; };

	/*
	** OK!
	*/
	::CloseServiceHandle( schService );
	return 1;

start_service_fail:
	::CloseServiceHandle( schService );
	return -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_stopService( const char *pServiceName )
{
int iRunning;
	assert( pServiceName );

	SC_HANDLE scH = Internal_openServiceManager();
	if ( !scH )
		{ return -1; };

	HANDLE schService = OpenService(
		scH, pServiceName, SERVICE_STOP  | SERVICE_QUERY_STATUS );

	Internal_closeServiceManager( scH );
			
	if ( !schService )
		{ return -1; };

	/*
	** Stop service
	*/
	SERVICE_STATUS ssStatus;
	if ( !ControlService( schService, SERVICE_CONTROL_STOP, &ssStatus ) )
		{ goto stop_service_fail; };

	/*
	** Track service while it stops
	*/
//	if ( Internal_trackServiceState( schService )!=0 )
//		{ goto stop_service_fail; };
iRunning = 1;
while( iRunning )
{
switch( Internal_trackServiceState( schService ) )
	{
	case 0: /* stopped */ 
			iRunning = 0;
			break;
	case 1: /* still running */				
			Sleep( 1000 );
			break;
	default: /* error */
			goto stop_service_fail;
	};
};

	/*
	** Ok!
	*/
	CloseServiceHandle( schService );
	return 1;

stop_service_fail:
	CloseServiceHandle( schService );
	return -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_isPi3Running( const char *pPi3NamedSemaphore )
{
	assert( pPi3NamedSemaphore );

	/*
	** Try to open the semaphore
	*/
	HANDLE hSema = ::OpenSemaphore( SEMAPHORE_ALL_ACCESS, FALSE, 
		pPi3NamedSemaphore );

	if ( !hSema )
		{
		/*
		** Semaphore does not exist, process is not running
		*/
		return 0;
		}

	CloseHandle( hSema );

	/*
	** Process is running
	*/
	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_startPi3( const char *pCommandLine,
		const char * /* pPi3NamedSemaphore */ )
{
	assert( pCommandLine );

	STARTUPINFO tStartUp;
	memset(&tStartUp, 0, sizeof(tStartUp));
	tStartUp.cb=sizeof(tStartUp);

	PROCESS_INFORMATION tProcessInfo;

	/* --- start the program --- */
	BOOL bProcessCreated = CreateProcess(
		NULL,				/* address of module name */
		(char *)pCommandLine,	/* address of command line */
		NULL,				/* address of process security attributes */
		NULL,				/* address of thread security attributes */
		FALSE,				/* new process inherits handles */
		0,					/* creation flags */
		NULL,				/* address of new environment block */
		NULL,				/* address of current directory name */
		&tStartUp,			/* address of STARTUPINFO */
		&tProcessInfo		/* address of PROCESS_INFORMATION */
		);

	if ( bProcessCreated )
		{
		::CloseHandle( tProcessInfo.hThread );
		::CloseHandle( tProcessInfo.hProcess );

		return 1;
		};

	return -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_stopPi3( const char *pPi3NamedSemaphore )
{
	assert( pPi3NamedSemaphore );

	/*
	** Try to open the semaphore
	*/
	HANDLE hSema = ::OpenSemaphore( SEMAPHORE_ALL_ACCESS, FALSE, 
		pPi3NamedSemaphore );

	if ( !hSema )
		{
		/*
		** Semaphore does not exist, process is not running
		*/
		return -1;
		};

	/*
	** Release the semaphore
	*/
	BOOL bRet = ::ReleaseSemaphore( hSema, 1, NULL );

	CloseHandle( hSema );

	return bRet ? 1 : -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void CTRL_semaphoreNameFromConfigPath( const char *pConfig,
	PIString &sSemaphore )
{
	/*
	** Semaphore name for application is the full
	** path to the configuration file in upper case
	*/
	PIString sSemaphoreName;
	enum { BUF_SIZE=1023 };
	char szBuf[BUF_SIZE+1];
	::GetCurrentDirectory( BUF_SIZE, szBuf );
	RelativeToAbsolutePath( szBuf, pConfig, sSemaphoreName );
	sSemaphoreName.ConvertToUpperCase();
	/*
	** Replace '\\' with '/'. Semaphore names cannot have '\\' in
	** them, see CreateSemaphore(), though this does not seem to be
	** a problem on Windows 95.
	*/
	SwapSubStrings sSwap( sSemaphoreName, "\\", "/" );	
	sSemaphore = sSwap.GetString();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Returns 1 on success, 0 on failure.
\*____________________________________________________________________________*/
int CTRL_openURL( HWND hWnd, const char *pURL )
{
	assert( pURL );

	if ( (int)::ShellExecute( hWnd, NULL, pURL, NULL,
		NULL, SW_SHOWNORMAL)<=32 )
		{ return 0; };

	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_isStartupInstalled( const char *pSemaphoreName )
{
	assert( pSemaphoreName);

	HKEY hKey;

	LONG lRet = ::RegQueryValueEx(
		HKEY_LOCAL_MACHINE,				// handle of key to query 
		REGSTR_PATH_RUNSERVICES,		// address of name of value to query 
		0,								// reserved 
		0,								// address of buffer for value type 
		0,								// address of data buffer 
		0								// address of data buffer size 
		);

#if 0
	switch( lRet )
		{
		case ERROR_SUCCESS: break;
		case ERROR_FILE_NOT_FOUND: return 0;
		default:
			return -1;
		};

	if ( ::RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		REGSTR_PATH_RUNSERVICES,
		0, KEY_ALL_ACCESS,
		&hKey )!=ERROR_SUCCESS )
		{ return -1; };
#endif

	lRet = ::RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		REGSTR_PATH_RUNSERVICES,
		0, KEY_ALL_ACCESS,
		&hKey );

	switch( lRet )	
		{
		case ERROR_SUCCESS: break;
		case ERROR_FILE_NOT_FOUND: return 0;
		default:
			return -1;
		};

	int iRet = (::RegQueryValueEx( hKey, pSemaphoreName, 0, 0, 0, 0 )==
		ERROR_SUCCESS ) ? 1 : 0;

	::RegCloseKey( hKey );
	
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_installStartup( const char *pSemaphoreName, const char *pCmdLine )
{
	assert( pSemaphoreName && pCmdLine );

	HKEY hKey = 0;
	DWORD dwDisp;

	if ( ::RegCreateKeyEx(
		HKEY_LOCAL_MACHINE,				// handle of an open key 
		REGSTR_PATH_RUNSERVICES,		// address of subkey name 
		0,								// reserved 
		"",								// address of class string 
		0,								// special options flag 
		KEY_ALL_ACCESS,					// desired security access 
		NULL,							// address of key security structure 
		&hKey,							// address of buffer for opened handle  
		&dwDisp							// address of disposition value buffer
		)!=ERROR_SUCCESS )
		{ goto an_error; };
	
	if ( ::RegSetValueEx(
		hKey,							// handle of key to set value for  
		pSemaphoreName,					// address of value to set 
		0,								// reserved
		REG_SZ,							// flag for value type 
		(const unsigned char *)pCmdLine,// address of value data
		1								// size of value data 
		)!=ERROR_SUCCESS )
		{ goto an_error; };	

	::RegCloseKey( hKey );

	return 1;

an_error:
	if ( hKey ) { ::RegCloseKey( hKey ); };

	return -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int CTRL_unInstallStartup( const char *pSemaphoreName )
{
	assert( pSemaphoreName );

	HKEY hKey;

	if ( ::RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		REGSTR_PATH_RUNSERVICES,
		0, KEY_ALL_ACCESS,
		&hKey )!=ERROR_SUCCESS )
		{ return -1; };

	int iRet = (::RegDeleteValue(hKey, pSemaphoreName )==ERROR_SUCCESS ) ?
		1 : -1;

	::RegCloseKey( hKey );

	return iRet;
}

