/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/DirExist.cpp,v $
 * $Date: 2002/01/26 20:34:00 $
 *
 Description:
	Handler for checking a directory exists.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"
#include "PIStrStr.h"
#include "DblList.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_INDEXFILE		"IndexFile"

#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	DirectoryExists

Description:
	Executes the following steps:
<UL>
<LI>If the physical path does not have a trailing '/' character then
cause a '301 Permanent Redirect' response to the client with the corrected
URL. This correctative action avoids some problems with directory indexing.
<LI>If the directory contains one of the files specified by the 'IndexFile'
parameter then cause an internal redirect to that file.
</UL>

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IndexFile
<TD>-
<TD>Filename
<TD>Index file name
<TD>IndexFile="index.html"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	IndexFile
</H5>
Specify the name of a directory index file to be sent to the client
instead of an automatically generated index. This parameter can be
repeated multiple time to specify alternate indexfiles. Indexfiles are
used in the order they are specified.

Phase:
	CHECKPATH

Returns:
	PIAPI_COMPLETED if the phase completed succesfully. PIAPI_CONTINUE
	if no action was taken, this will occur when the specified resource
	is not a directory. PIAPI_ERROR and PIAPI_ABORT respectively when
	generic and severe errors occur.

Example:
	<PRE>
	&lt;Object&gt;
		Name DirectoryExists
		Class DirectoryExistsClass
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		CheckPath DirectoryExists IndexFile="index.html" \
			IndexFile="default.html"
		...
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class DirectoryExists : public HandlerBaseHTTP
{
private:
	/* --- forbid copy constructor --- */
	DirectoryExists( const DirectoryExists &t )
	: HandlerBaseHTTP( t )
		{ assert( 0 ); };

	/* ---
	Configuration data
	--- */
	DblList lIndexFileNames;

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "DirectoryExists: ";
		if ( !PIUtil_stricmp( KEY_CONF_INDEXFILE, pVariable ) )
			{
			lIndexFileNames.Append( (DblList::type)PI_NEW(PIString( pValue ) ));
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	DirectoryExists( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject )
		{
		ReadParameters( iArgc, ppArgv );
		};

	virtual ~DirectoryExists()
		{
		for( DblListIterator i(lIndexFileNames); !i.BadIndex(); i++)
			{
			PI_DELETE( (PIString *)i.Current() );
			};
		};

	virtual int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer & )
		{
		assert( IsOK() );
		if ( iPhase!=PH_CHECKPATH )
			{ return PIAPI_ERROR; };

		PIDB *pQ = tPIHTTP.pRequestDB;
		PIDB *pR = tPIHTTP.pResponseDB;
		PIDB *pC = tPIHTTP.pConnectionDB;
		const char *pPath = (const char *)PIDB_lookup( pR,
			PIDBTYPE_STRING, KEY_INT_PATH, 0 );
		PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath ); 

		if ( !PIFInfo_exists( pFInfo ) )
			{
			HTTPCore_releaseCachedFile( pFInfo );
			return PIAPI_CONTINUE;
			};

		if ( !PIFInfo_isDirectory( pFInfo ) )
			{
			HTTPCore_releaseCachedFile( pFInfo );
			return PIAPI_CONTINUE;
			};

		/* --- no longer need to keep the file object --- */
		HTTPCore_releaseCachedFile( pFInfo );

		/* ---
		If the directory does not end in '/', then issue a 
		permanent redirect to correct this, otherwise directory
		indexing will not function correctly
		--- */
		int iLen = strlen( pPath );
		if ( iLen==0 ||
			( pPath[iLen-1]!=HTTP_DIRSEPERATOR 
#if WIN32
			&& pPath[iLen-1]!='\\'
#endif
			) )
			{

			/* ---
			Generate the new 'Location'
			Change the URI to what it should be and redirect
			--- */
			const char *pURI = (const char *)PIDB_lookup( pQ,
				PIDBTYPE_STRING, KEY_HTTP_URI, 0 );
			PIOStrStream ostr;
			if ( PIDB_lookup( pC, PIDBTYPE_STRING, KEY_HTTPS_KEYSIZE, 0 ) )
				{
				/*
				** Running under SSL
				*/
				ostr << "https://";
				}
			else
				{
				/*
				** Not running under SSL
				*/
				ostr << "http://";
				};
			const char *pHostName = HTTPUtil_getHostName( &tPIHTTP );
			if ( pHostName )
				{ ostr << pHostName; };
			const char *pPort = HTTPUtil_getHostPort( &tPIHTTP );
			if ( pPort )
				{ ostr << ':' << pPort; };

			if ( pURI ) 
				{ ostr << pURI; };

			ostr << HTTP_DIRSEPERATOR;
			ostr << ends;
			PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_LOCATION,
				(void *)ostr.str(), 0 );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_PERMANENTREDIRECT );
			};

		/* ---
		The path is a directory and it has a trailing '/'.
		Does an indexfile exist?
		--- */
		
		/* ---
		loop over every possible index file
		--- */
		for( DblListIterator i(lIndexFileNames); !i.BadIndex(); i++ )
			{
			PIString &sIndexFile = *( (PIString *)i.Current() );
			PIString sPath( pPath );
			sPath.Concatenate( sIndexFile );
			PIFInfo *pTmpFileInfo = HTTPCore_getCachedFile( sPath );

			/* --- test if the file exists --- */
			int iExists = ( pTmpFileInfo && PIFInfo_exists( pTmpFileInfo ));

			if ( iExists )
				{
				/* --- reset the path to the index file --- */
				PIDB_replace( pR, PIDBTYPE_STRING, KEY_INT_PATH,
					(void *)(const char *)sPath, 0 );

				HTTPCore_releaseCachedFile( pTmpFileInfo );
				return PIAPI_COMPLETED;
				};
			HTTPCore_releaseCachedFile( pTmpFileInfo );

			};

		return PIAPI_COMPLETED;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int DirectoryExists_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( DirectoryExists( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name DirectoryExistsClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor DirectoryExists_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name DirectoryExists
		Class DirectoryExistsClass
	</Object>

/*___+++CNF_END+++___*/
#endif

