/*____________________________________________________________________________*\
 *

 Copyright (c) 1999 Holger Zimmermann, John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/ISAPIFLT/ISAPIFLT.cpp,v $
 * $Date: 2002/07/25 13:41:35 $
 *
 Description:
	Load and execute an ISAPI filter.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <iostream.h>
#include <stdio.h>

#if WIN32
//#include <HTTPExt.h>
#include <windows.h>
#include <HTTPFilt.h>
#endif

#include "PIString.h"
#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPDefs.h"
#include "HTTPUtil.h"
#include "MiscUtil.h"
#include "PIStrStr.h"
#include "PIUtil.h"
#include "DeQuote.h"
#include "StrToken.h"
#include "Base64.h"

#if WIN32

/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	ISAPIFLT

Description:
	Load and execute an ISAPI (Internet Server API) filter DLL. You can find more
	informations about ISAPI filters in the <A HREF="http://msdn.microsoft.com/library/psdk/iisref/isgu0q0n.htm">
	MSDN library</A>. This handler will load the configured ISAPI filter DLL when
	the server is started. If you want to load more than one ISAPI filter you must
	use one ISAPIFLT handler for each filter DLL. During a HTTP request the handler
	will dispatch the registered ISAPI filter notifications to the filter DLL.

Options:
<H4>Overview</H4>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>Condition
<TD>-
<TD>A Pi3Expression
<TD>If condition is TRUE, filter is responsible for the request
<TD>Condition "&regexp('*/admin/*',$z)"

<TR>
<TD>FilterDLLPath
<TD>+
<TD>A valid file path
<TD>The path to an ISAPI filter DLL
<TD>FilterDLLPath "C:\Pi3Web\Isapi\daf.dll"

<TR>
<TD>Terminators
<TD>-
<TD>Init, Headers, Hostmap, Mapping, CheckPath, CheckAuth,
CheckAccess, CheckType, Handle, Log, Destroy
<TD>Phases, that are terminated by this handler
<TD>Terminators Headers | Mapping | Log

<TR>
<TD>SystemLogon
<TD>Yes
<TD>Yes|No
<TD>Indicates if the server should attempt a system logon
<TD>Terminators Headers | Mapping | Log

<TR>
<TD>Variable
<TD>-
<TD>&lt;Pi3Expression&gt;
<TD>A variable definition
<TD>Variable "GATEWAY_INTERFACE=CGI/1.1"

<TR>
<TD>ExtraHeaders
<TD>Yes
<TD>Yes|No
<TD>Indicates if extra headers are considered
<TD>ExtraHeaders Yes

<TR>
<TD>ExtraHeadersPrefix
<TD>-
<TD>&lt;A string&gt;
<TD>Used as prefix of each extra header
<TD>ExtraHeadersPrefix &quot;HTTP_&quot;

<TR>
<TD>ExtraHeadersIgnore
<TD>-
<TD>&lt;Space delimited Strings&gt;
<TD>List of unconsidered extra headers
<TD>ExtraHeadersIgnore &quot;Content-Type Content-Length&quot;

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>Condition</H5>
If the condition given by the Pi3Expression is TRUE, the filter controlled
by this handler is responsible for the request. The filter will get all
notifications registered with GetFilterVersion during startup. This gives
the ISAPI filter handlers more flexibility and a superior control mechanism
as replacement for the IIS filter priority chain.

<H5>FilterDLLPath</H5>
The DLL given by this option will be loaded during server startup and
controlled by the handler during HTTP requests. Only ISAPI filter DLLs
are allowed here. If the DLL is not an ISAPI filter or the filter version
is not compliant or the filter uses an unsupported notification type the
DLL will not be loaded and the server start will be aborted with an
error message.

<H5>Terminators</H5>
Pi3Web server needs exactly one handler to complete one HTTP request phase.
For efficiency the following handlers will not be called if a phase is
completed. From this it follows sometimes we have to change the default
termination of a phase (i.e. Mapping ) to give ISAPI filters a chance to
be called by the main dispatcher. Sometimes we have to complete a phase by
an ISAPI filter (i.e. a Log filter). If the filter DLL returns
"SF_STATUS_REQ_HANDLED_NOTIFICATION", PIAPI_COMPLETE will be returned
automatically by the handler. Otherwise we can use this option to explicite
complete a phase if the filter is called. All phase names are allowed here
separated by "|".

<H5>SystemLogon</H5>
If this parameter is switched off and the returned status by the filter
is SF_STATUS_REQ_HANDLED_NOTIFICATION, this is treated as successful
authentication.
If this parameter is switched on and the returned status by the filter
is SF_STATUS_REQ_HANDLED_NOTIFICATION, the Pi3Web will attempt a system
logon.

<H5>Variable</H5>
Specifies a server variable expression to be set into the buffer
of ISAPI extension calls to GetServerVariable. Could be set multiple
times in one configuration object.

<H5>ExtraHeaders</H5>
This configuration key defines, if extra headers are considered by
the processing of this handler.

<H5>ExtraHeadersPrefix</H5>
This configuration value will be the trailing string of each extra
request header.

<H5>ExtraHeadersIgnore</H5>
The headers in this list are not treated as extra headers.

Phase:
	All Phases

Returns:
	PIAPI_COMPLETED, PIAPI_CONTINUE, PIAPI_ERROR, or INT_REDIRECT according
	to the status returned by the filter.

Notes:
<H5>General</H5>
	The ISAPI filter handler is an experimental plug in. Since the IIS and
	Pi3Web has different architectures it is difficult to be 100% compliant.

<H5>Parameters</H5>
	The Pi3Expressions in the configuration variables may contain besides
	the standard shortcuts used in Pi3Expressions the following, context
	specific parameters.

	<CENTER>
	<TABLE BORDER=1>
	<TH>Parameter<TH>Evaluates to<TR>
	<TD>%a<TD>All extra headers, as required for header ALL_HTTP<TR>
	<TD>%l<TD>Content-Length as DWORD, as required for ISAPI<TR>
	</TABLE>
	</CENTER>

<H5>Filter version</H5>
	The version of the filter DLL is checked against the server ISAPI version
	(2.1). If the version isn't compliant the filter isn't loaded and the server
	start will abort with an error message.

<H5>Filter notifications</H5>
	All filter notification events are supported except of SF_NOTIFY_READ_RAW_DATA
	and SF_NOTIFY_SEND_RAW_DATA events. If the filter try to register these events,
	the filter will not be loaded and the server start will abort with an error message.

<H5>Notification order</H5>
	Filter priority is not supported with Pi3Web. If you chain ISAPI filters with
	other	handlers the order is fixed in Config.Pi3 in the "Handlers" section of the
	HTTPLogicObject. To much filters with lots of notifications will noticeable slow
	down the server.

<H5>Phase completion</H5>
	An architecture sign of Pi3Web is, that one handler must finish a connection
	phase	by returning PIAPI_COMPLETED. If the filter DLL will return
	"SF_STATUS_REQ_HANDLED_NOTIFICATION" after handling a notification, the handler
	will return PIAPI_COMPLETED and the phase is finished. If the filter does not,
	then you can force this with the handler option "Terminators".

<H5>ServerSupportFunction</H5>
	The following support functions are not supported:
	<P>SF_REQ_SET_NEXT_READ_SIZE
	<P>SF_REQ_GET_CONNID

<H5>Server variables</H5>
	A pre-defined set of IIS server variables are supported. "ALL_HTTP" will
	hold all request headers except of the values stored in other server variables.
	This is very flexible by adding, changing or removing "Variable" parameters
	from the configuration file.

<H5>Authentication filters</H5>
	If an authentication filter returns SF_STATUS_REQ_HANDLED_NOTIFICATION on
	notification SF_NOTIFY_AUTHENTICATION, the server will attempt to logon and
	impersonate the user at the system dependent on setting "SystemLogon".
	<P>This works only on Windows NT/2000/XP and requires permissions to
	"act as part of the operating system" for the user running the Pi3Web
	(except XP). Otherwise the HTTP statuscode 401 and appropriate response
	headers to force a basic authentication	are set by the Pi3Web.

Example:
	<PRE>
	&lt;Object&gt;
		Name DAFFLT
		Class ISAPIFLTClass
		Condition "&regexp('*/admin/*',$z)"
		FilterDLLPath "C:\Pi3Web\Isapi\daf.dll"
		Terminators Headers | Mapping
	&lt;/Object&gt;
	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Prototypes for callback functions
	Class for ISAPI context
\*____________________________________________________________________________*/
BOOL WINAPI fnGetServerVariable( PHTTP_FILTER_CONTEXT pfc,
					LPSTR	 lpszVariableName,
					LPVOID	 lpvBuffer,
					LPDWORD	 lpdwSize );

BOOL WINAPI fnWriteClient( PHTTP_FILTER_CONTEXT pfc,
					LPVOID	 Buffer,
					LPDWORD	 lpdwBytes,
					DWORD	 dwReserved );

VOID * WINAPI fnAllocMem ( PHTTP_FILTER_CONTEXT pfc,
					DWORD    cbSize,
					DWORD    dwReserved );

BOOL WINAPI fnServerSupportFunction ( PHTTP_FILTER_CONTEXT pfc,
					enum SF_REQ_TYPE sfReq,
					PVOID pData,
					DWORD ul1,
					DWORD ul2 );

BOOL WINAPI fnAddResponseHeaders ( PHTTP_FILTER_CONTEXT pfc,
					LPSTR lpszHeaders,
					DWORD dwReserved );

BOOL WINAPI fnGetHeader ( PHTTP_FILTER_CONTEXT pfc,
					LPSTR   lpszName,
					LPVOID  lpvBuffer,
					LPDWORD lpdwSize );

BOOL WINAPI fnSetHeader ( PHTTP_FILTER_CONTEXT pfc,
					LPSTR  lpszName,
					LPSTR  lpszValue );

BOOL WINAPI fnAddHeader ( PHTTP_FILTER_CONTEXT pfc,
					LPSTR  lpszName,
					LPSTR  lpszValue );

typedef BOOL  (WINAPI * PFN_GETFILTERVERSION)( HTTP_FILTER_VERSION *pVer );
typedef DWORD (WINAPI * PFN_HTTPFILTERPROC )( PHTTP_FILTER_CONTEXT pfc,
											  DWORD NotificationType,
											  VOID *pvNotification );

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_INT_ISAPICONTEXT "ISAPI_CONTEXT"
#define KEY_CONF_CONDITION "Condition"
#define KEY_CONF_FLTDLLPATH "FilterDLLPath"
#define KEY_CONF_TERMINATORS "Terminators"
#define KEY_CONF_EXTRAHDRPREFIX					"ExtraHeadersPrefix"
#define KEY_CONF_EXTRAHEADERSIGNORE				"ExtraHeadersIgnore"
#define KEY_CONF_EXTRAHEADERS					"ExtraHeaders"
#define KEY_CONF_SYSTEMLOGON					"SystemLogon"
#define VALUE_NO								"No"
#define VALUE_YES								"Yes"
#define KEY_CONF_VARIABLE						"Variable"

/*____________________________________________________________________________*\
 *
 Flags and Options
\*____________________________________________________________________________*/
#define FLG_NONE		0x00000000
#define FLG_INIT		0x00000001
#define FLG_HEADERS		0x00000002
#define FLG_HOSTMAP		0x00000004
#define FLG_MAPPING		0x00000008
#define FLG_CHECKPATH	0x00000010
#define FLG_CHECKAUTH	0x00000020
#define FLG_CHECKACCESS	0x00000040
#define FLG_CHECKTYPE	0x00000080
#define FLG_HANDLE		0x00000100
#define FLG_LOG			0x00000200
#define FLG_DESTROY		0x00000400

/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
				Map these flags to names

 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
struct {
	const char *pName;
	int iFlag;
} aFlagMap[] =
	{
	{ "Init",				FLG_INIT },
	{ "Headers",			FLG_HEADERS },
	{ "Hostmap",			FLG_HOSTMAP },
	{ "Mapping",			FLG_MAPPING },
	{ "CheckPath",			FLG_CHECKPATH },
	{ "CheckAuth",			FLG_CHECKAUTH },
	{ "CheckAccess",		FLG_CHECKACCESS },
	{ "CheckType",			FLG_CHECKTYPE },
	{ "Handle",				FLG_HANDLE },
	{ "Log",				FLG_LOG },
	{ "Destroy",			FLG_DESTROY },	/* ---

	/* --- leave this last always --- */
	{ 0, 0 }
	};

/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
			Map these phase flags to phases by index

 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
int aPhaseMap[] =
	{
//	{ FLG_NONE },
	{ FLG_INIT },
	{ FLG_HEADERS },
	{ FLG_HOSTMAP },
	{ FLG_MAPPING },
	{ FLG_CHECKPATH },
	{ FLG_CHECKAUTH },
	{ FLG_CHECKACCESS },
	{ FLG_CHECKTYPE },
	{ FLG_HANDLE },
	{ FLG_LOG },
	{ FLG_DESTROY },	/* ---

	/* --- leave this last always --- */
	{ 0 }
	};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Class for ISAPI context
\*____________________________________________________________________________*/
class ISAPIFLT;
class ISAPIContext : public _HTTP_FILTER_CONTEXT
{
public:
	ISAPIFLT *pISAPIFLT;
	PIHTTP *pPIHTTP;
	HTTP_FILTER_VERSION tVer;
	const char *pszHeaders;
	int iRc;
	HANDLE hToken;
	DWORD cbTotalBytes;

	ISAPIContext( ISAPIFLT *pTheISAPIFLT, PIHTTP *pThePIHTTP,
		HTTP_FILTER_VERSION tTheVer )
	:	pISAPIFLT( pTheISAPIFLT ),
		pPIHTTP( pThePIHTTP ),
		tVer( tTheVer ),
		pszHeaders( NULL),
		iRc( PIAPI_COMPLETED ),
		hToken( 0 )
		{
		/* --- initialize member data --- */
		cbSize = sizeof( _HTTP_FILTER_CONTEXT );
		Revision = HTTP_FILTER_REVISION;
		ServerContext = (PVOID)this;

		/* --- a filter function may initialize this --- */
		pFilterContext = NULL;

		/* --- callback functions --- */
		GetServerVariable = fnGetServerVariable;
		AddResponseHeaders = fnAddResponseHeaders;
		WriteClient = fnWriteClient;
		AllocMem = fnAllocMem;
		ServerSupportFunction = fnServerSupportFunction;

		/* --- proprietary data --- */
		const char *pContentLength = (const char *)PIDB_lookup(
			pPIHTTP->pRequestDB, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, 0 );
		cbTotalBytes = pContentLength ? atol( pContentLength ) : 0;
		fIsSecurePort = (BOOL)PIDB_lookup( pPIHTTP->pConnectionDB,
			PIDBTYPE_STRING, KEY_HTTPS_KEYSIZE, 0 );
		};

	~ISAPIContext()
		{
		CloseHandle( hToken );
		}
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ISAPIFLT : public HandlerBase
{
private:
	/* ---
	Configuration data
	--- */
	PIDB *pTypes;					/* DB for variable types */
	PIFInfo *pFInfo;				/* Store DLL file info */
	PIDLL *pDLL;					/* Store DLL handle */
	const char *pPath;              /* Store DLL path for error msgs */
	Pi3Expression *pCondition;		/* Store condition */
	HTTP_FILTER_VERSION tVer;		/* The server version of ISAPI filter */
	PFN_HTTPFILTERPROC fnHttpFilterProc; /* Pointer to the client filter function */
	const char *pfkIsapiContext;	/* Lookup key for the context pointer */
	const char *pFKPath;			/* Lookup key for the path */
	int iResult;					/* Result code for Handler

	/* ---
	Configuration data
	--- */
   	PIString sExtraHeadersPrefix;	/* Extra Header for environment values, */
									/* by default "HTTP_" */
	int iExtraHeaders;				/* whether or not to send extra headers */
	DblList lIgnoreHeaders;			/* extra headers to ignore */
	int iSystemLogon;
	FnPi3Write aFunctions[256];		/* callback functions */
	int iOptions;					/* The options of this handler */

	inline int checkIgnoreList(const char *pKey)
		{
		/* --- exclude the variables in lIgnoreHeaders --- */
		const char *pExclude;
		for( DblListIterator l( lIgnoreHeaders ); !l.BadIndex(); l++ )
			{
			pExclude = (const char *)l.Current();
			if (!PIUtil_stricmp(pKey, pExclude))
				{
				return 1;
				}
			};
		return 0;
		}
protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os.flags( os.flags() | ios::hex );
		os << pWhere << "ISAPIFLT: ";
		if ( !PIUtil_stricmp( KEY_CONF_CONDITION, pVariable ))
			{
			Pi3String *pError = Pi3String_new( 0 );
			pCondition = Pi3Expression_new( pValue, 0, pError );
			PIString sError( Pi3String_getPtr( pError ) );
			Pi3String_delete( pError );
			if ( !pCondition )
				{
				os << "Syntax error in condition: " << sError << ends;
				CONFIG_ERR( Object(), os.str() );
				Pi3Expression_delete( pCondition );
				return 0;
				};
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXTRAHDRPREFIX, pVariable ) )
			{
				sExtraHeadersPrefix = (const char *)DeQuote( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXTRAHEADERSIGNORE, pVariable ) )
			{
			StringTokenizer tTokens( (const char *)DeQuote(pValue), " " );
			for( int i=0; i<tTokens.NumTokens(); i++)
				{
				lIgnoreHeaders.Append( (DblList::type)
					PI_NEW( PIString( tTokens.GetToken( i ) ) ) );
				};
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXTRAHEADERS, pVariable ) )
			{
			if ( !PIUtil_stricmp( VALUE_NO, pValue ) )
				{ iExtraHeaders = 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_SYSTEMLOGON, pVariable ) )
			{
			if ( !PIUtil_stricmp( VALUE_NO, pValue ) )
				{ iSystemLogon = 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_VARIABLE, pVariable ) )
			{
			const char *pK;
			const char *pV;
			StringTokenizer tTokens( (const char *)DeQuote(pValue), "=" );
			if (tTokens.NumTokens() < 2)
				{
				os << "Invalid Variable '" << pValue <<
					"'" << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				}

			pK = tTokens.GetToken( 0 );
			pV = tTokens.GetToken( 1 );

			Pi3String *pError = Pi3String_new( 0 );
			Pi3Expression *pExpr = Pi3Expression_new( pV, aFunctions,
				pError );
			if ( !pExpr )
				{
				os << Pi3String_getPtr( pError ) << ends;
				Pi3String_delete( pError );
				CONFIG_ERR( Object(), os.str() ); 
				return 0;
				};
			Pi3String_delete( pError );
			PIDB_add( pTypes, PIDBTYPE_OPAQUE, pK, (void *)pExpr, 0 );
			}

        else if ( !PIUtil_stricmp( KEY_CONF_TERMINATORS, pVariable ))
			{
			/* ---
			Split the options up by the bar ('|') character and then
			strip trailing and leading whitespace from option before
			attempting to match it.
			--- */
			StringTokenizer tTokens( pValue, "|" );
			for(int i=0; i<tTokens.NumTokens(); i++)
				{
				const char *pToken = tTokens.GetToken( i );

				/* --- skip leading whitespace --- */
				for( ; *pToken && (isspace(*pToken)); pToken++ );

				/* ---
				include only to first whitespace trailing whitespace
				--- */
				int j=0;
				for( ; pToken[j] && !(isspace(pToken[j])); j++ );
				/* ---
				j now contains the length of the first word in pToken
				--- */

				/* ---
				cycle through the list of available flags
				comparing them with this one
				--- */
				int i=0;
				for( ; aFlagMap[i].pName; i++ )
					{
					if ( !PIUtil_strncmpi( aFlagMap[i].pName, pToken, j ) )
						{ break; };
					};
				if ( !aFlagMap[i].pName )
					{
					/* --- flag not found --- */
					PIString sTmp( pToken, j );
					os << "Unknown option flag '" << sTmp << "'." << ends;
					CONFIG_ERR( Object(), os.str() );
					return 0;
					};
				iOptions |= aFlagMap[i].iFlag;
				};	/* --- loop over tokens separated by '|' --- */
			}
        else if ( !PIUtil_stricmp( KEY_CONF_FLTDLLPATH, pVariable ))
			{

			PIString sPath( pValue );
			HTTPCore_relativeToAbsolutePath(PIObject_getDB( Object() ),
				pValue,	(int *)&sPath );

			pFInfo = HTTPCore_getCachedFile( sPath ); 
			if ( !pFInfo )
				{ 
				os << "ISAPI Filter DLL '" << pValue << \
					"' does not exist or is not a regular file." << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};

			/* ---	
			This file must exist and be a regular file 
			--- */
			if ( !PIFInfo_exists( pFInfo ) || !PIFInfo_isRegular( pFInfo ) )
				{
				os << "ISAPI Filter DLL '" << pValue << \
					"' does not exist or is not a regular file." << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};

			/* ---	
			Try to load the ISAPI filter DLL 
			--- */
			strcpy((char *)pPath, sPath );
			pDLL = PIDLL_new( pPath );
			if ( !PIDLL_isLoaded( pDLL ) )
				{
				const char *pDesc = PIDLL_getErrorDescription( pDLL );
				os << "ISAPI Filter DLL '" << pValue << \
					"' could not be loaded, error description is " << pDesc << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};

			/* ---
			Load the GetFilterVersion function from the DLL
			--- */
			PFN_GETFILTERVERSION fnGetVersion = (PFN_GETFILTERVERSION)
				PIDLL_getAddress( pDLL, "GetFilterVersion" );
			if ( !fnGetVersion )
				{
				os << "Could not load function 'GetFilterVersion'" << \
					  " from ISAPI DLL '" << pPath << "'" << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};

			/* ---
			Do GetFilterVersion
			--- */
			tVer.dwServerFilterVersion = HTTP_FILTER_REVISION;
			if ( !( (fnGetVersion)( &tVer ) ) )
				{
				os << "Could not execute function 'GetFilterVersion'" << \
				" from ISAPI DLL '" << pPath << "'. Win32 Error code is " << \
				::GetLastError() << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				}
			/* ---
			Is the Version of the filter backward compatible to our server module?
			--- */
			else if ( tVer.dwServerFilterVersion < tVer.dwFilterVersion)
				{
				os << "Incompatible filter version in ISAPI DLL '" << pPath << \
					  "'. Server Filter version is " << tVer.dwServerFilterVersion << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
 				}
			/* ---
			This server module hasn't support for read/write filter notifications.
			If the filter needs this, you couldn't use the filter with Pi3Web
			--- */
			else if ( !( tVer.dwFlags & ( SF_NOTIFY_READ_RAW_DATA | SF_NOTIFY_SEND_RAW_DATA )))
				{
				/* ---
				Load the filter function
				--- */
				fnHttpFilterProc = ( PFN_HTTPFILTERPROC )PIDLL_getAddress( pDLL, "HttpFilterProc" );
				if ( !fnHttpFilterProc )
					{
					os << "Could not load 'HttpFilterProc' from ISAPI DLL '" << pPath << "'" << ends;
					CONFIG_ERR( Object(), os.str() );
					return 0;
					};
				}
			else
				{
				os << "Pi3Web doesn't support raw data read/write notification " << \
					  "requested by ISAPI DLL '" << pPath << "'" << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};
			}
		else
			{
			os << "Unknown directive '" << pVariable <<	"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		return 1;
		};

	/* ---
	Callback function for Pi3Expression parameter %a
	--- */
	static int f_a( PIHTTP *pPIHTTP, void *pData, char *pszBuffer, int iLength )
	{
		(void)pPIHTTP;
		ISAPIContext &tData = *( (ISAPIContext *)pData );
	    assert( pszBuffer );

		PIDB *pQ = pPIHTTP->pRequestDB;											
		int i = 0;																
	
		/* --- loop over all RFC822 headers in request block --- */				
		PIDBIterator *pIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );	
		if ( !pIter ) { return 0; }

		const char *pKey;													
		const char *pValue;
		for(; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter ) )
			{																
			pValue = (const char *)PIDBIterator_current( pIter,	&pKey );
			assert( pKey && pValue );											

			if ( !pKey || !pValue ) { /* sanity */ continue; };										

			if (tData.pISAPIFLT->checkIgnoreList(pKey)) continue;

			void *fnEnv = PIDB_lookup( tData.pISAPIFLT->pTypes, PIDBTYPE_OPAQUE, pKey, 0 );
			if ( fnEnv ) { continue; };

			int iVarLen = strlen( pKey );										
			int iVarPos = tData.pISAPIFLT->sExtraHeadersPrefix.Len();								
			int iValSize = strlen( pValue );									

			// keep space for ': \r\n\0'
			if ( (i+iVarPos+iVarLen+iValSize+5) < iLength )							
				{																
				memcpy( &(pszBuffer[i]), tData.pISAPIFLT->sExtraHeadersPrefix, iVarPos );
				i += iVarPos;
				for( int j=0; j<iVarLen; j++, i++ )							
					{																
					if ( isalnum( pKey[j] ) )										
						{ pszBuffer[i]=toupper(pKey[j]); }						
					else															
						{ pszBuffer[i]='_'; };									
					};
				pszBuffer[i++] = ':';
				pszBuffer[i++] = ' ';
				memcpy( &(pszBuffer[i]), pValue, iValSize ); i+=iValSize;		
				pszBuffer[i++] = '\r';
				pszBuffer[i++] = '\n';											
				}																
			else																
				{																
				PIDBIterator_delete( pIter );
				::SetLastError( ERROR_INSUFFICIENT_BUFFER );				
				return 1024;													
				};																
			};
		PIDBIterator_delete( pIter );
		pszBuffer[i] = 0;											
		return i;															
	};

	/* ---
	Callback function for Pi3Expression parameter %l
	--- */
	static int f_l( PIHTTP *pPIHTTP, void *pData, char *pszBuffer, int iLength )
	{
	(void)pPIHTTP;
	ISAPIContext &tData = *( (ISAPIContext *)pData );
    assert( pszBuffer );

	*((DWORD *)pszBuffer) = tData.cbTotalBytes;
	pszBuffer[sizeof(DWORD)] = 0;
	return sizeof(DWORD);
	}

public:
	ISAPIFLT( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBase( pObject ),
		pTypes( PIDB_new( 0, "Types" ) ),
		pFInfo( NULL ),
		pDLL( NULL ),
		fnHttpFilterProc( NULL ),
		pPath( (const char *)PIUtil_malloc( 256 ) ),
		pCondition( NULL ),
		iOptions( 0 ),
		pfkIsapiContext( PIDB_getFastKey( KEY_INT_ISAPICONTEXT, PIDBTYPE_STRING )),
		pFKPath( PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING )),
		iResult( PIAPI_CONTINUE ),
		iExtraHeaders( 1 ),
		iSystemLogon( 1 )
		{
		/* --- clear out callback functions --- */
		memset( aFunctions, 0, sizeof( FnPi3Write ) * 256 );
		aFunctions['a'] = f_a;
		aFunctions['l'] = f_l;
		ReadParameters( iArgc, ppArgv );
		};

	virtual ~ISAPIFLT()
		{
		PIDB_delete( pTypes );
		for( DblListIterator i( lIgnoreHeaders ); !i.BadIndex(); i++ )
			{ PI_DELETE( (PIString *)i.Current() ); };
		Pi3Expression_delete( pCondition );
		if (pDLL) PIDLL_delete( pDLL );
		if (pFInfo) HTTPCore_releaseCachedFile( pFInfo );
		PIUtil_free( (void *)pPath );
		};

	/* ---
	The real filter stuff
	--- */
	int RunISAPI( PIHTTP *pPIHTTP, int iPhase, ISAPIContext *pContext )
		{
		assert( pPIHTTP && fnHttpFilterProc && pContext);
		int iFilterRet = 0;

		/* ---
		Look what we've to do
		--- */
		DWORD nType = 0;
		LPVOID pvNotification = NULL;

		/* ---
		Nothing to do: if (non)secure port notification is other
		than port characteristics of the connection
		--- */
		if (( pContext->fIsSecurePort && !( pContext->tVer.dwFlags & SF_NOTIFY_SECURE_PORT ))
		|| !( pContext->fIsSecurePort || ( pContext->tVer.dwFlags & SF_NOTIFY_NONSECURE_PORT )))
			{
			return FALSE;
			};

		/* ---
		Preprocessing headers phase mapped to SF_NOTIFY_PREPROC_HEADERS
		--- */
		if ( !nType && ( iPhase == PH_HEADERS ) && ( pContext->tVer.dwFlags & SF_NOTIFY_PREPROC_HEADERS ))
			{
			nType = SF_NOTIFY_PREPROC_HEADERS;
			HTTP_FILTER_PREPROC_HEADERS Hdrs;
			pvNotification = &Hdrs;
			Hdrs.GetHeader = fnGetHeader;
			Hdrs.SetHeader = fnSetHeader;
			Hdrs.AddHeader = fnAddHeader;

			/* ---
			Not implemented in filter version 2.1
			--- */
//			Hdrs.HttpStatus = pPIHTTP->iStatus;
			};

		/* ---
		Mapping phase mapped to SF_NOTIFY_URL_MAP
		--- */
		if ( !nType && ( iPhase == PH_MAPPING ) && ( pContext->tVer.dwFlags & SF_NOTIFY_URL_MAP ))
			{
			nType = SF_NOTIFY_URL_MAP;
			HTTP_FILTER_URL_MAP urlMap;
			pvNotification = &urlMap;

			urlMap.pszURL = (const char *)PIDB_lookup( pPIHTTP->pRequestDB,
				PIDBTYPE_STRING, KEY_HTTP_URI, 0 );

			urlMap.pszPhysicalPath = (char *)PIDB_lookup( pPIHTTP->pResponseDB,
				PIDBTYPE_STRING, pFKPath, PIDBFLAG_FASTKEY );
			CHAR lpvBuffer[256];
			urlMap.cbPathBuff = sizeof(lpvBuffer);
			strncpy( lpvBuffer, urlMap.pszPhysicalPath, strlen(urlMap.pszPhysicalPath) + 1);
			urlMap.pszPhysicalPath = lpvBuffer;
		};

		/* ---
		Authentication phase mapped to SF_NOTIFY_AUTHENTICATION
		--- */
		if ( !nType && ( iPhase == PH_CHECKAUTH ) && ( pContext->tVer.dwFlags & SF_NOTIFY_AUTHENTICATION )
			// don't notify internal redirects again!
			&& ( pPIHTTP->iStatus != ST_UNAUTHORIZED ))
			{
			CHAR lpvUserBuff[SF_MAX_USERNAME] = "";
			CHAR lpvPasswBuff[SF_MAX_PASSWORD] = "";

			/*
			** Get browser authentication string
			*/
			const char *pAuthenticate = (const char *)PIDB_lookup(
				pPIHTTP->pRequestDB, PIDBTYPE_RFC822, KEY_HTTP_AUTHORIZATION, 0 );
			if ( pAuthenticate )
				{

				/*
				** Check authentication string
				*/
				if ( !strncmp( pAuthenticate, "Basic ", 6 ) )
					{
					/*
					** Check to see if this is an acceptable base64 encoding
					*/
					const char *pBase64 = &( pAuthenticate[6] );
					CHAR auth[256];

					Base64Decode( pBase64, auth, sizeof( auth ));
					char *tok = strtok( auth, ":" );
					if ( tok != NULL )
						{
						strncpy( lpvUserBuff, tok, strlen(tok) + 1);
						PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING,
							KEY_INT_REMOTEUSER, tok, 0 );
						tok = strtok( NULL, ":" );
						if ( tok != NULL )
							{
							strncpy( lpvPasswBuff, tok, strlen(tok) + 1);
							PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING,
								KEY_INT_AUTHPASS, tok, 0 );
							};
						};
					};
				};

			nType = SF_NOTIFY_AUTHENTICATION;
			HTTP_FILTER_AUTHENT Auth;
			pvNotification = &Auth;
			Auth.cbUserBuff = sizeof(lpvUserBuff);
			Auth.pszUser = lpvUserBuff;
			Auth.cbPasswordBuff = sizeof(lpvPasswBuff);
			Auth.pszPassword = lpvPasswBuff;
		};

		/* ---
		CheckPath phase mapped to SF_NOTIFY_ACCESS_DENIED if HTTP state is 401 || 403
		--- */
		if ( !nType && ( iPhase == PH_CHECKPATH ) && (( pPIHTTP->iStatus == ST_UNAUTHORIZED ) || ( pPIHTTP->iStatus == ST_FORBIDDEN ))
			&& ( pContext->tVer.dwFlags & SF_NOTIFY_ACCESS_DENIED ))
			{
			nType = SF_NOTIFY_ACCESS_DENIED;
			HTTP_FILTER_ACCESS_DENIED Accs;
			pvNotification = &Accs;
			Accs.pszURL = (const char *)PIDB_lookup( pPIHTTP->pRequestDB,
				PIDBTYPE_STRING, KEY_HTTP_URI, 0 );

			/*
			** if Status is 403, the original path is stored in KEY_INT_OLDPATH
			*/
			if ( pPIHTTP->iStatus == ST_FORBIDDEN )
				{
				PIDB_getFastKey( KEY_INT_OLDPATH, PIDBTYPE_STRING );
				}
			else
				{
				PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING );
				};

			Accs.pszPhysicalPath = (const char *)PIDB_lookup( pPIHTTP->pResponseDB,
				PIDBTYPE_STRING, pFKPath, PIDBFLAG_FASTKEY );

			Accs.dwReason = ( pPIHTTP->iStatus == ST_UNAUTHORIZED ) ? SF_DENIED_LOGON : SF_DENIED_RESOURCE;
			};

		/* ---
		Handle phase mapped to SF_NOTIFY_SEND_RESPONSE
		--- */

		/* Not implemented in filter version 2.1
		if ( !nType && ( iPhase == PH_HANDLE ) && ( pContext->tVer.dwFlags & SF_NOTIFY_SEND_RESPONSE ))
			{
			nType = SF_NOTIFY_SEND_RESPONSE;
			HTTP_FILTER_SEND_RESPONSE Resp;
			pvNotification = &Resp;
			Resp.GetHeader = fnGetHeader;
			Resp.SetHeader = fnSetHeader;
			Resp.AddHeader = fnAddHeader;

			Resp.HttpStatus = pPIHTTP->iStatus;
			};
		*/

		/* ---
		Log phase mapped to	SF_NOTIFY_LOG
		--- */
		if ( !nType && ( iPhase == PH_LOG ) && ( pContext->tVer.dwFlags & SF_NOTIFY_LOG ))
			{
			nType = SF_NOTIFY_LOG;
			HTTP_FILTER_LOG Log;
			pvNotification = &Log;
			Log.pszClientHostName = (const char *)PIDB_lookup( pPIHTTP->pConnectionDB, PIDBTYPE_STRING, KEY_INT_REMOTEHOST, 0 );
			Log.pszClientUserName = (const char *)PIDB_lookup( pPIHTTP->pResponseDB, PIDBTYPE_STRING, KEY_INT_REMOTEUSER, 0 );
			Log.pszServerName = (const char *)PIDB_lookup( pPIHTTP->pHostDB, PIDBTYPE_RFC822, KEY_INT_HOSTNAME, 0 );
			Log.pszOperation = (const char *)PIDB_lookup( pPIHTTP->pRequestDB, PIDBTYPE_STRING, KEY_HTTP_METHOD, 0 );
			Log.pszTarget = (const char *)PIDB_lookup( pPIHTTP->pResponseDB, PIDBTYPE_STRING, KEY_INT_SCRIPTNAME, 0 );
			Log.pszParameters = (const char *)PIDB_lookup( pPIHTTP->pRequestDB, PIDBTYPE_STRING, KEY_HTTP_QUERYSTRING, 0 );
			Log.dwHttpStatus = pPIHTTP->iStatus;
			Log.dwWin32Status = ::GetLastError();
			/* ---
			Not implemented in filter version 2.1
			--- */
//			Log.dwBytesSent = pPIHTTP->GetIOBuffer().GetBytesSent();
//			Log.dwBytesRecvd = 0;
//			long lSeconds;	
//			long lMicroSeconds;	
//			PITimer_elapsed( pPIHTTP->pTimer, &lSeconds, &lMicroSeconds );
//			Log.msTimeForProcessing = ( 1000 * lSeconds ) + (lMicroSeconds % 1000 ));
			};

		/* ---
		Nothing to do!
		--- */
		if ( !nType ) return FALSE;

		/* ---
		Fire off filter function
		--- */
		iFilterRet = (fnHttpFilterProc)( pContext, nType, pvNotification );

		/* ---
		Post processing for handled notifications
		--- */

		/*
		** Change the physical path for phase PH_MAPPING
		*/
		if (( iPhase == PH_MAPPING ) && ( pContext->tVer.dwFlags & SF_NOTIFY_URL_MAP )
			&& ( iFilterRet != SF_STATUS_REQ_ERROR ))
			{
			const char *pFKPath = PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING );
			PIDB_replace( pPIHTTP->pResponseDB,	PIDBTYPE_STRING, pFKPath,
				((HTTP_FILTER_URL_MAP *)pvNotification)->pszPhysicalPath,
				PIDBFLAG_FASTKEY );
			};

		/*
		** Change the authentication uid/pwd in the DB
		*/
		if (( iPhase == PH_CHECKAUTH ) && ( pContext->tVer.dwFlags & SF_NOTIFY_AUTHENTICATION ))
			{
			if (( iFilterRet != SF_STATUS_REQ_ERROR ) && ( (DWORD)::GetLastError() != ERROR_ACCESS_DENIED ))
				{
				PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING, KEY_INT_REMOTEUSER,
					((HTTP_FILTER_AUTHENT *)pvNotification)->pszUser, 0 );
				PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING, KEY_INT_AUTHPASS,
					((HTTP_FILTER_AUTHENT *)pvNotification)->pszPassword, 0 );

				/*
				** If we have finished, try to logon (WindowsNT only)
				*/
				if ( iFilterRet == SF_STATUS_REQ_HANDLED_NOTIFICATION )
					{
					/*
					** parameter not set or already authenticated
					*/
					if (!iSystemLogon || pContext->hToken)
						{
						iResult = PIAPI_COMPLETED;
						return pContext->iRc;
						};

					OSVERSIONINFO VerInfo;
					VerInfo.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
					// If this fails, something has gone wrong
					if ( !GetVersionEx( &VerInfo )) { return PIAPI_ERROR; }

					if ( VerInfo.dwPlatformId == VER_PLATFORM_WIN32_NT )
						{
						if (LogonUser(((HTTP_FILTER_AUTHENT *)pvNotification)->pszUser,
							NULL, // lpszDomain
							((HTTP_FILTER_AUTHENT *)pvNotification)->pszPassword,
							LOGON32_LOGON_INTERACTIVE,
							LOGON32_PROVIDER_DEFAULT, 
							&pContext->hToken ))
							{
							// If this fails, something has gone wrong
							if (!ImpersonateLoggedOnUser( pContext->hToken ))
								{
								return PIAPI_ERROR;
								}
							/*
							** authenticated and impersonated
							*/
							iResult = PIAPI_COMPLETED;
							return pContext->iRc;
  							}
						}
					}
				}
			else
				{
				if ((DWORD)::GetLastError() != ERROR_ACCESS_DENIED ) { return PIAPI_ERROR; }
				}

			/*
			** authentication challenge (Basic)
			*/
			PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_STRING, KEY_INT_AUTHTYPE, "Basic", 0 );
			const char *pRealm = (const char *)PIDB_lookup( pPIHTTP->pResponseDB,
				PIDBTYPE_STRING, KEY_INT_AUTHENTICATIONREALM, 0 );
			char *pAuth = (char *)PIHTTP_allocMem( pPIHTTP, sizeof( "Basic realm=\"\"" )
				+ strlen( pRealm ) + 1 );
			sprintf( pAuth, "Basic realm=\"%s\"", pRealm );
			PIDB_replace( pPIHTTP->pResponseDB, PIDBTYPE_RFC822,
				KEY_HTTP_WWW_AUTHENTICATE, pAuth, 0 );

			/*
			** Redirect because of authentication failure
			*/
			iResult = HTTPUtil_doHTTPError( pPIHTTP, ST_UNAUTHORIZED );
			return pContext->iRc;
			};

		/*
		** if ServerSupportFunction added any extra header
		** requested with SF_REQ_ADD_HEADERS_ON_DENIAL
		*/
		if (( iPhase == PH_CHECKPATH ) && (( pPIHTTP->iStatus == ST_UNAUTHORIZED )
			|| ( pPIHTTP->iStatus == ST_FORBIDDEN )) && pContext->pszHeaders )
			{
			char *pszValue;
			char *pszName = strtok( (char *)pContext->pszHeaders, ":" );
			while ( pszName )
				{
				// search for space but also remove CRLF
				pszValue = strtok( NULL, "\r\n ");
				if ( !(pszName && pszValue) )
					{
					break;
					};
				PIDB_add( pPIHTTP->pResponseDB, PIDBTYPE_RFC822, pszName, pszValue, 0 );
				// search for double quota but also remove CRLF
				pszName = strtok( NULL, "\r\n:" );
				};
			};

		/* ---
		Status/Error handling
		--- */
		switch( iFilterRet )
			{
			case SF_STATUS_REQ_FINISHED:
				/*
				** Assume Content-Length could not be set, disable keep open
				*/
				PIDB_replace( pPIHTTP->pConnectionDB, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN, 0, 0 );
				pPIHTTP->iStatus = ST_OK;
				break;

			case SF_STATUS_REQ_FINISHED_KEEP_CONN:
				/*
				** OK
				*/
				pPIHTTP->iStatus = ST_OK;
				break;

			case SF_STATUS_REQ_NEXT_NOTIFICATION:
				/*
				** OK
				*/
				break;

			case SF_STATUS_REQ_HANDLED_NOTIFICATION:
				/*
				** OK
				*/
				iResult = PIAPI_COMPLETED;
				break;

			case SF_STATUS_REQ_READ_NEXT:
				/*
				** OK
				*/
				break;

			case SF_STATUS_REQ_ERROR:
			default:
				pPIHTTP->iStatus = ST_INTERNALERROR;
				HTTPCore_logError( pPIHTTP, "ISAPIFLT: ISAPI filter \
'%s' returned SF_STATUS_REQ_ERROR or an unknown return code.", pPath );
			};

		return pContext->iRc;
		};


	/* ---
	The entry point ...
	--- */
	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &tB/* tBuffer */ )
		{
        /*
        ** The result for this phase depends on preconfigured values
        */
		iResult = ( aPhaseMap[iPhase] & iOptions ) ? PIAPI_COMPLETED : PIAPI_CONTINUE;

        /*
        ** Only handle the request if the condition is TRUE
        */
		if ( pCondition && !Pi3Expression_write( pCondition, &tPIHTTP, 0, 0, 0 )) return iResult;

		int iFilterErr = FALSE;
		/*
		** Retrieve ISAPI context ...
		*/
		ISAPIContext *pContext = (ISAPIContext *)PIDB_lookup( tPIHTTP.pConnectionDB,
			PIDBTYPE_OPAQUE, pfkIsapiContext, PIDBFLAG_FASTKEY );
		/*
		** ... or create a new one
		*/
		if ( !pContext )
			{
			pContext = new ISAPIContext( this, &tPIHTTP, tVer );
			PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE, pfkIsapiContext,
				(void *)pContext, PIDBFLAG_FASTKEY );
			};

        /*
        ** Only handle the request if the current status is a non-error status
        */
        if (( !tPIHTTP.iStatus ) || ( tPIHTTP.iStatus == ST_OK )
		/* --- Exception: 401, 403 are notified by ISAPI filters! --- */
		|| ( tPIHTTP.iStatus == ST_UNAUTHORIZED ) || ( tPIHTTP.iStatus == ST_FORBIDDEN ))
            {
			/* --- OK, do the filter DLL processing --- */
			iFilterErr = RunISAPI( &tPIHTTP, iPhase, pContext );
			};

		/*
		** pContext - explicite destruction
		*/
		if (( iPhase == PH_DESTROY ) && pContext )
			{
			delete pContext;
			pContext = NULL;
			PIDB_replace( tPIHTTP.pConnectionDB, PIDBTYPE_OPAQUE, pfkIsapiContext,
				(void *)pContext, PIDBFLAG_FASTKEY);
			};

		/*
		** an error is occured
		*/
		if ( iFilterErr )
			{
			HTTPCore_logError( &tPIHTTP, "ISAPIFLT: Call to 'HttpFilterProc' \
failed for ISAPI DLL '%s'. Win32 Error code is %d.", pPath , ::GetLastError() );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			}
		else return iResult;
		};


	/* ---
	Used by the callback function with the same name
	--- */
	BOOL GetServerVariable( PIHTTP *pPIHTTP, LPSTR lpszVariableName,
		LPVOID lpvBuffer, LPDWORD lpdwSize, ISAPIContext *pContext )
		{
		if ( !lpszVariableName || !lpdwSize )
			{
			::SetLastError( ERROR_INVALID_PARAMETER );
			return FALSE;
			};

		/* ---
		Convert variable to upper case
		--- */
		PIString sTmp( lpszVariableName );
		sTmp.ConvertToUpperCase();

		Pi3Expression *pExpr = (Pi3Expression *)PIDB_lookup(pTypes, PIDBTYPE_OPAQUE,
			(const char *)sTmp, 0);

		if (pExpr)
			{
			int iWritten = Pi3Expression_write( pExpr, pPIHTTP, pContext, (char *)lpvBuffer,
				*lpdwSize );
			if ( iWritten==-1 )
				{
				*lpdwSize = 0;
				::SetLastError( ERROR_INVALID_INDEX );
				return 0;
				}

			/*
			** Skip empty variable fields, this allows an environment
			** variable to be added conditionally
			*/
			if ( iWritten>=0 )
				{
				if ( iWritten>=*lpdwSize )
					{
					*lpdwSize = iWritten;
					::SetLastError( ERROR_INSUFFICIENT_BUFFER );
					return 0;
					};
				((char *)lpvBuffer)[iWritten] = '\0';	
				*lpdwSize = iWritten;
				return 1;
				};
			}

		/* --- no variable found, try to find an extra header --- */
		if ( iExtraHeaders ) 
			{
			PIDB *pQ = pPIHTTP->pRequestDB;											
	
			/* --- loop over all RFC822 headers in request block --- */				
			PIDBIterator *pIter = PIDB_getIterator( pQ, PIDBTYPE_RFC822, 0, 0 );	
			if ( !pIter ) { return 0; }

			const char *pKey;													
			const char *pValue;
			for(; PIDBIterator_atValidElement( pIter ); PIDBIterator_next( pIter ) )
				{
				pValue = (const char *)PIDBIterator_current( pIter,	&pKey );
				assert( pKey && pValue );											

				if ( !pKey || !pValue ) { /* sanity */ continue; };

				if (checkIgnoreList(pKey)) continue;

				PIString sTest(sExtraHeadersPrefix);
				sTest.Concatenate(pKey);
				char *pBuf = (char *)(const char *)sTest;
				for( int i=0; i<sTest.Len(); i++ )							
					{																
					if ( isalnum( pBuf[i] ) )										
						{ pBuf[i]=toupper(pBuf[i]); }						
					else															
						{ pBuf[i]='_'; };		
					}

				if (!PIUtil_stricmp((const char *)sTest, (const char *)sTmp))
					{
					strncpy((char *)lpvBuffer, pValue, strlen(pValue));
					*lpdwSize = strlen(pValue);
					((char *)lpvBuffer)[*lpdwSize] = '\0';	
					return 1;
					}
				}
			PIDBIterator_delete( pIter );

			*(char *)lpvBuffer = '\0';
			*lpdwSize = 0;
			::SetLastError( ERROR_INVALID_INDEX );
			return 0;
			}
		
		::SetLastError( ERROR_INVALID_INDEX );
		return 0;
	}
};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL WINAPI fnGetServerVariable( PHTTP_FILTER_CONTEXT pfc,
					LPSTR lpszVariableName,
					LPVOID  lpvBuffer,
					LPDWORD lpdwSize )
{
	ISAPIContext *pContext = (ISAPIContext *)pfc;
	BOOL iRet = pContext->pISAPIFLT->GetServerVariable( pContext->pPIHTTP,
		lpszVariableName, lpvBuffer, lpdwSize, pContext );
	if ( HTTPCore_debugEnabled() )
		{
		HTTPCore_logDebug( DBG_MED, "ISAPI20: Return code<%d>, variable<%s>, \
value<%x>, size<%d>.",
			iRet, lpszVariableName ? lpszVariableName : "",
			(void *)lpvBuffer, lpdwSize ? *lpdwSize : 0 );
		};
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL WINAPI fnWriteClient( PHTTP_FILTER_CONTEXT pfc,
					LPVOID  Buffer,
					LPDWORD lpdwBytes,
					DWORD   dwReserved )
{
	if ( !pfc || !pfc->ServerContext || !lpdwBytes || !Buffer )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return FALSE;
		};

	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	/*
	** This function does not yet exist!
	*/
//	pContext->pISAPIFLT->notifyWrite( Buffer, lpdwBytes );
	int iWritten = PIIOBuffer_write( pContext->pPIHTTP->pBuffer,
		(char *)Buffer, *lpdwBytes, PIIOBUF_NONE );
	if ( iWritten == -1 )
		{
		/*
		** Socket must have been closed
		*/
		::SetLastError( ERROR_INVALID_HANDLE );
		return FALSE;
		};
	*lpdwBytes = iWritten;
	return TRUE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
VOID * WINAPI fnAllocMem ( struct _HTTP_FILTER_CONTEXT * pfc,
					DWORD    cbSize,
					DWORD    dwReserved )
{
	if ( !pfc || !pfc->ServerContext )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return NULL;
		};
	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	return PIHTTP_allocMem( pContext->pPIHTTP, cbSize ) ;
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL WINAPI fnServerSupportFunction( PHTTP_FILTER_CONTEXT pfc,
					enum SF_REQ_TYPE sfReq,
					PVOID pData,
					DWORD ul1,
					DWORD ul2 )
{
	if ( !pfc || !pfc->ServerContext )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return FALSE;
		};

	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	PIHTTP *pPIHTTP = pContext->pPIHTTP;
	const char *pFKProxyProtocol = PIDB_getFastKey( KEY_HTTP_PROXYPROTOCOL,	PIDBTYPE_STRING );

	switch( sfReq )
		{
		case SF_REQ_ADD_HEADERS_ON_DENIAL:
			pContext->pszHeaders = (const char *)pData;
			return TRUE;

		/*
		** Not implemented in filter version 2.1
		*/
//		case SF_REQ_DISABLE_NOTIFICATIONS:
//			pContext->tVer.dwFlags &= (ul1 ^ 0xFFFFFFFF); 
//			return TRUE;

		case SF_REQ_SEND_RESPONSE_HEADER:
			{
			int iStatus = pData ? atoi( (char *)pData ) : ST_OK;	
			pPIHTTP->iStatus = iStatus;
			pContext->iRc = PIAPI_COMPLETED;

			/*
			** Send server specific headers
			*/
			HTTPCore_sendGeneralHeaders( pPIHTTP );
			if ( pData )
				{
				DWORD dwSize = strlen( (const char *)pData );
				PIIOBuffer_write( pPIHTTP->pBuffer, (const char *)pData,
					dwSize, PIIOBUF_NONE );
				};
			if ( ul1 )
				{
				DWORD dwSize = strlen( (const char *)ul1 );
				PIIOBuffer_write( pPIHTTP->pBuffer, (const char *)ul1,
					dwSize, PIIOBUF_NONE );
				};
			};
			return TRUE;

		/*
		** Not supported!
		*/
		case SF_REQ_SET_NEXT_READ_SIZE:
			ul1 = 0;
			return TRUE;

		case SF_REQ_SET_PROXY_INFO:
			ul1 = (DWORD)PIDB_lookup( pPIHTTP->pRequestDB, PIDBTYPE_STRING,
				pFKProxyProtocol, PIDBFLAG_FASTKEY );
			return TRUE;

		/*
		** Not supported!
		*/
		case SF_REQ_GET_CONNID:
			*(DWORD *)pData = 0;
			return TRUE;
		default:;
		};
	return FALSE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
BOOL WINAPI fnAddResponseHeaders ( PHTTP_FILTER_CONTEXT pfc,
					LPSTR lpszHeaders,
					DWORD dwReserved )
{
	if ( !pfc || !pfc->ServerContext || !lpszHeaders )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return FALSE;
		};

	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	PIHTTP *pPIHTTP = pContext->pPIHTTP;

	char *pszValue;
	char *pszName = strtok( (char *)lpszHeaders, ":" );
	while ( pszName )
		{
		/*
		** search for space but also remove CRLF
		*/
		pszValue = strtok( NULL, "\r\n ");
		if ( !(pszName && pszValue) )
			{
			break;
			};
		PIDB_add( pPIHTTP->pResponseDB, PIDBTYPE_RFC822, pszName, pszValue, 0 );
		/*
		** search for double quota but also remove CRLF
		*/
		pszName = strtok( NULL, "\r\n:" );
		};
	return TRUE;
};


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
//
//  Retrieves the specified header value.  Header names should include
//  the trailing ':'.  The special values 'method', 'url' and 'version'
//  can be used to retrieve the individual portions of the request line
//
BOOL WINAPI fnGetHeader (	PHTTP_FILTER_CONTEXT pfc,
							LPSTR   lpszName,
							LPVOID  lpvBuffer,
							LPDWORD lpdwSize )
{

# define GET_STR( db, type, key )                                   \
{																	\
	const char *pValue = (const char *)PIDB_lookup( db,	type, key, 0 );	\
	if ( !pValue ) { pValue = ""; };								\
	int iToCopy = strlen( pValue );									\
	if (( iToCopy > *lpdwSize ) || !lpvBuffer )						\
		{															\
		::SetLastError( ERROR_INSUFFICIENT_BUFFER );				\
		return FALSE;												\
		};															\
	*lpdwSize = iToCopy;											\
	strncpy( (char *)lpvBuffer, pValue, iToCopy + 1);				\
	return TRUE;													\
};			

	if ( !pfc || !pfc->ServerContext || !lpszName )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return FALSE;
		};

	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	PIHTTP *pPIHTTP = pContext->pPIHTTP;

	if ( !strcmp(lpszName, "method" ))
		{
		GET_STR( pPIHTTP->pRequestDB, PIDBTYPE_STRING, KEY_HTTP_METHOD );
		}
	else if ( !strcmp( lpszName, "url" ))
		{
		GET_STR( pPIHTTP->pRequestDB, PIDBTYPE_STRING, KEY_HTTP_URI );
		}
	else if ( !strcmp(lpszName, "version" ))
		{
		GET_STR( pPIHTTP->pRequestDB, PIDBTYPE_STRING, KEY_HTTP_PROTOCOL );
		}
	else 
		{
		if ( strrchr( lpszName, ':' )) { *strrchr( lpszName, ':' ) = '\0'; };
		GET_STR( pPIHTTP->pRequestDB, PIDBTYPE_RFC822, strupr(lpszName) );
		};

	return TRUE;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
//
//  Replaces this header value to the specified value.  To delete a header,
//  specified a value of '\0'.
//
BOOL WINAPI fnSetHeader (	PHTTP_FILTER_CONTEXT pfc,
							LPSTR  lpszName,
							LPSTR  lpszValue )
{
	if ( !pfc || !pfc->ServerContext || !lpszName )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return FALSE;
		};

	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	PIHTTP *pPIHTTP = pContext->pPIHTTP;

	if ( strrchr( lpszName, ':' )) { *strrchr( lpszName, ':' ) = '\0'; };
	return PIDB_replace( pPIHTTP->pRequestDB, PIDBTYPE_RFC822, strupr(lpszName), lpszValue, 0 );
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Adds the specified header and value to the response
\*____________________________________________________________________________*/

BOOL WINAPI fnAddHeader (	PHTTP_FILTER_CONTEXT pfc,
							LPSTR  lpszName,
							LPSTR  lpszValue )
{
		if ( !pfc || !pfc->ServerContext || !lpszName )
		{
		::SetLastError( ERROR_INVALID_PARAMETER );
		return FALSE;
		};

	ISAPIContext *pContext = (ISAPIContext *)pfc->ServerContext;
	PIHTTP *pPIHTTP = pContext->pPIHTTP;

	if ( strrchr( lpszName, ':' )) { *strrchr( lpszName, ':' ) = '\0'; };
	return PIDB_add( pPIHTTP->pResponseDB, PIDBTYPE_RFC822, lpszName, lpszValue, 0 );
}

#endif /* WIN32 */

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int ISAPIFLT_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
#if WIN32
	return HandlerBase_constructor( pObj, PI_NEW( ISAPIFLT( pObj,
		iArgc, ppArgv ) ) );
#elif POSIX
	(void)pObj;
	(void)iArgc;
	(void)ppArgv;
	return PIAPI_COMPLETED;
#endif
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int ISAPIFLT_destructor( PIObject *pObj, int,
	const char *[] )
{
#if WIN32
	delete (ISAPIFLT *)PIObject_getUserData( pObj );
#endif
	return PIAPI_COMPLETED;
}


#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name ISAPIFLTClass
		Type LogicExtension
		Library ISAPIFLT
		OnClassLoad HandlerBaseISAPIFLT_onClassLoad
		Constructor ISAPIFLT_constructor
		CopyConstructor HandlerBaseISAPIFLT_copyConstructor
		Destructor ISAPIFLT_destructor
		Execute HandlerBaseISAPIFLT_execute
	</Class>

	<Object>
		Name ISAPIFLT
		Class ISAPIFLTClass
	</Object>

/*___+++CNF_END+++___*/
#endif

