/*____________________________________________________________________________*\

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann.  All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/SSL/SSL.c,v $
 * $Date: 2002/07/25 16:32:22 $
 *
 Description:

	SSL layer based on OpenSSL (http://www.openssl.org), formerly known as
	Eric Youngs SSLeay.

	The SSLeay BIO architecture is very similar to the PIIO object 
	server loadable class. They both provide an abstract interface to
	an IO layer than can be a terminal transport or a filter layer to
	another IO object. This makes them both pretty easy to integrate.
	
	The transport IO object is created and wrapped in an SSLeay BIO
	object. This is then used to create an SSLeay SSL object which
	in turn is made to look like a PIIO object so other components in pi3
	can use it as an IO object.

	Fragments of source code in this file were copied from Eric Youngs
	SSL distribution. His copyright statement is included along with
	this file, its called COPYRIGHT.eay

	TODO:
		- Test client setup.
		- Remove SSL_debug stuff.

\*____________________________________________________________________________*/
/* $SourceTop:$ */

/*
** Top level flags
*/
#define PISSL_DEBUG	0	/* development time flag */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "ssl.h"		/* From SSLeay */
#include "err.h"		/* From SSLeay */
#include "rand.h"		/* From SSLeay */

#define INIT_ERR \
	{ \
	ERR_print_errors_fp(stderr); \
	if ( g_pStdioErr ) { ERR_print_errors_fp(g_pStdioErr); }; \
	}
#define RUNTIME_ERR(x) \
	{ \
	if ( g_pStdioErr ) { ERR_print_errors_fp(g_pStdioErr); }; \
	fprintf(stderr,"SSI[%s, %d]: %s\n", __FILE__, __LINE__, (x) ); \
	}

#if PISSL_DEBUG
	enum { s_quiet = 0 };
#	define D { fprintf(stderr,"%s: %d\n", __FILE__, __LINE__ ); }
#else
	enum { s_quiet = 1 };
#	define D
#endif

#include "PiAPI.h"
#include "Pi2API.h"

/*____________________________________________________________________________*\
 Description:
	Documentation for this IO object.
\*____________________________________________________________________________*/

#if 0
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	SSL

Description:
	Perform SSL (secure sockets layer) on a pi3 IO channel. An instance
	of this IO class would normally be inserted between a TCPIP transport
	IO object and a higher level logic component implementing a TCP
	service to make the service secure, e.g. (HTTP-->HTTPS). 
<P>
	However this IO filter can be used on any IO chain, both for
	other transports (shared memory, named pipes) and other uses, FastCGI
	connections, FTP connections etc.
<P>
	Typically an instance of this object is inserted into the IO chain
	using code like the following.
<PRE>
	Substitute:

	...
	IOObject TheTransportIO 
	...

	With:
	
	...
	IOObject SSL IOObject="TheTransportIO"
	...

</PRE>

	This SSL pi3 component is implemented using the OpenSSL library (
	formerly known as Eric Youngs SSLeay library, thus -
<PRE>
   &quot;This product includes software developed by the OpenSSL Project
	for use in the OpenSSL Toolkit (http://www.openssl.org/).&quot;
</PRE>

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IOObject
<TD>+
<TD>&lt;pi3objectname&gt;
<TD>A Pi3 object
<TD>IOObject="TCPIPIO"

<TR>
<TD>Type
<TD>Passive
<TD>Active|Passive
<TD>Specify IO layer type
<TD>Type="Passive"

<TR>
<TD>Version
<TD>23
<TD>2, 3, 23 or 10
<TD>Version of SSL/TLS to accept
<TD>Version=3

<TR>
<TD>Verify
<TD>0
<TD>0, 1 or 2
<TD>Client certificate verification requirements
<TD>Verify=2

<TR>
<TD>VerifyDepth
<TD>8
<TD>An integer value
<TD>Level to verify peer certificate
<TD>VerifyDepth=2

<TR>
<TD>CACertificatePath
<TD>-
<TD>&lt;pathname&gt;
<TD>CA Path 
<TD>CACertificatePath="./certs"

<TR>
<TD>CACertificateFile
<TD>-
<TD>&lt;filename&gt;
<TD>CA Certificate file to use
<TD>CACertificateFile="./ca.pem"

<TR>
<TD>PrivateKeyFile
<TD>+*
<TD>&lt;filename&gt;
<TD>A private key file
<TD>PrivateKeyFile="./key.pem"

<TR>
<TD>CertificateFile
<TD>+*
<TD>&lt;filename&gt;
<TD>Certificate file
<TD>CertificateFile="./cert.pem"

<TR>
<TD>Flag
<TD>-
<TD>"Bugs"
<TD>Flag
<TD>Flag="Bugs"

<TR>
<TD>DebugFile
<TD>-
<TD>"STDOUT", "STDERR" or &lt;filename&gt;
<TD>Specific an SSL debug file
<TD>DebugFile="./ssl.dbg"

<TR>
<TD>CipherList
<TD>-
<TD>
<TD>List of acceptable ciphers
<TD>CipherList "!ADH:RC4+RSA:HIGH:MEDIUM:LOW:EXP:+SSLv2:+EXP"
("DEFAULT", "NULL-MD5" - Read the openssl documentation for details)


</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>
<STRONG>+*</STRONG> in the <IT>default</IT> indicates
the field is mandatory for server connections only<BR>

<H4>Description of Options</H4>
<H5>IOObject</H5>
Lower level IO object which will be loaded to provide the transport for
this layer.

<H5>Type</H5>
Type of the IO filter layer, active (client) or passive (server).

<H5>Version</H5>
Version of SSL to accept, 2=SSLv2 only; 3=SSLv3 only; 23=SSLv2 or SSLv3; 10=TLS1.0

<H5>Verify</H5>
Client certificate verification mode. 
<UL>
<LI>0: No certificate verification is performed.
<LI>1: Certificate is verified, but not mandatory.
<LI>2: Certificate must be present.
</UL>

<H5>VerifyDepth</H5>
Depth to verify peer cetificate to before giving up.

<H5>CACertificatePath</H5>
CACertificate verification path.

<H5>CACertificateFile</H5>
CACertificate file.

<H5>PrivateKeyFile</H5>
Private keyfile.

<H5>CertificateFile</H5>
Certificate file.

<H5>Flag</H5>
Specify a flag which effects the behaviour of SSL. This directive
can be repeated multiple times to add different flags.
<UL>
<LI>Bugs: Turn on SSL bug compatibility.
<LI>Hacks: Turn on special SSLeay hacks.
<LI>Disable: Switch off SSL (Raw TCP connection).
</UL>

<H5>DebugFile</H5>
Produce an SSL debug file. Since the debug file is global to all
SSL IO object instances, the first SSL IO to specify a debug file
will determine the debug file created. If the filename is 'STDOUT'
or 'STDERR' the debugging will be output to the standard output
and standard error streams respectively.

<H5>CipherList</H5>
List of acceptable ciphers.

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>

	&lt;Object&gt;
		Name SSL
		Class SSLClass
		IOObject TCPIPIOObject
	&lt;/Object&gt;

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
 *
 Description:
	Prototypes.
\*____________________________________________________________________________*/
static int PIIO_SSLeay_write(BIO *h,char *buf,int num);
static int PIIO_SSLeay_read(BIO *h,char *buf,int size);
static int PIIO_SSLeay_puts(BIO *h,char *str);
static long PIIO_SSLeay_ctrl(BIO *h,int cmd,long arg1,char *arg2);
static int PIIO_SSLeay_new(BIO *h);
static int PIIO_SSLeay_free(BIO *data);

/*____________________________________________________________________________*\
 *
 Description:
	Data structures.
\*____________________________________________________________________________*/
struct _SSLConf
{
	/* ---
	
	Data global to all connections.

	Typically just configuration data.
	If any connections modify this data then mutex's must to used to ensure
	synchronization amoungst multiple threads
	--- */
	int iVersion;				/* SSL version to accept */
	int iVerify;				/* client verification to perform */
	int iVerifyDepth;			/* verification depth */
	char *pKeyF;				/* keyfile */
	char *pCertF;				/* certfile */
	char *pCACertP;				/* ca certificate path */
	char *pCACertF;				/* ca certificate file */
	char *pRandomSeedFile;		/* random seed file */
	unsigned int iRandomSize;	/* size of random seed */
	SSL_CTX *pCtx;				/* SSL_CTX */
	int iServer;				/* is this a server or client connection? */
	int iBugs;					/* SSL bug compatibility */
	int iHack;					/* not documented */
	int iDisable;				/* swith off SSL */
	char *pDebugFile;			/* debug file name */
	FILE *pStdioDebug;			/* stdio debug file */
	BIO *pBIODebug;				/* BIO debug file */
	char *pCipherList;			/* cipher list */
};
typedef struct _SSLConf SSLConf;

struct _SSL
{
	/* ---

	Per connection (SSL) data:

	Data unique to this connection. Since only one connection will
	modify the data at a time it may safely be used without thread
	synchronization
	--- */
	SSLConf *pConf;		/* pointer to configuration struct */
	int iIsPrototype;			/* true for prototype, false otherwise */
	PIObject *pIOObject;		/* IO object for this connection, or 
								** prototype object.
								*/
	int iError;					/* error flag */
	PIDB *pSaveDB;				/* save the DB */
	BIO *pBio;					/* BIO object */
	SSL *pSSL;					/* SSL structure */
};
typedef struct _SSL PISSL;

struct _CipherInformation
{
	/* ---

	Information on ciphers, key sizes, etc.

	--- */
	const char *pName;
	int iKeySize;
	int iSecretKeySize;
};
typedef struct _CipherInformation CipherInformation;
const CipherInformation aCiphers[]=
{
/*
** Stuff shared between SSLv2 and SSLv3
*/
{ SSL_TXT_NULL_WITH_MD5,					0,		0 },
{ SSL_TXT_RC4_128_WITH_MD5,					128,	128 },
{ SSL_TXT_RC4_128_EXPORT40_WITH_MD5,		128,	40 },
{ SSL_TXT_RC2_128_CBC_WITH_MD5,				128,	128 },
{ SSL_TXT_RC2_128_CBC_EXPORT40_WITH_MD5,	128,	40 },
{ SSL_TXT_IDEA_128_CBC_WITH_MD5,			128,	128 },
{ SSL_TXT_DES_64_CBC_WITH_MD5,				56,		56 },
{ SSL_TXT_DES_64_CBC_WITH_SHA,				56,		56 },
{ SSL_TXT_DES_192_EDE3_CBC_WITH_MD5,		168,	168 },
{ SSL_TXT_DES_192_EDE3_CBC_WITH_SHA,		168,	168 },
{ SSL_TXT_NULL,								0,		0 },


/*
** New with SSLv3
*/
{ SSL3_TXT_RSA_NULL_MD5,					0,		0 },
{ SSL3_TXT_RSA_NULL_SHA,					0,		0 },
{ SSL3_TXT_RSA_RC4_40_MD5,					128,	40 },
{ SSL3_TXT_RSA_RC4_128_MD5,					128,	128 },
{ SSL3_TXT_RSA_RC4_128_SHA,					128,	128 },
{ SSL3_TXT_RSA_RC2_40_MD5,					128,	40 },
{ SSL3_TXT_RSA_IDEA_128_SHA,				128,	128 },
{ SSL3_TXT_RSA_DES_40_CBC_SHA,				40,		40 },
{ SSL3_TXT_RSA_DES_64_CBC_SHA,				56,		56 },
{ SSL3_TXT_RSA_DES_192_CBC3_SHA,			168,	168 },
{ SSL3_TXT_DH_DSS_DES_40_CBC_SHA,			40,		40 },
{ SSL3_TXT_DH_DSS_DES_64_CBC_SHA,			56,		56 },
{ SSL3_TXT_DH_DSS_DES_192_CBC3_SHA,			168,	168 },
{ SSL3_TXT_DH_RSA_DES_40_CBC_SHA,			40,		40 },
{ SSL3_TXT_DH_RSA_DES_64_CBC_SHA,			56,		56 },
{ SSL3_TXT_DH_RSA_DES_192_CBC3_SHA,			168,	168 },
{ SSL3_TXT_EDH_DSS_DES_40_CBC_SHA,			40,		40 },
{ SSL3_TXT_EDH_DSS_DES_64_CBC_SHA,			56,		56 },
{ SSL3_TXT_EDH_DSS_DES_192_CBC3_SHA,		168,	168 },
{ SSL3_TXT_EDH_RSA_DES_40_CBC_SHA,			40,		40 },
{ SSL3_TXT_EDH_RSA_DES_64_CBC_SHA,			56,		56 },
{ SSL3_TXT_EDH_RSA_DES_192_CBC3_SHA,		168,	168 },
{ SSL3_TXT_ADH_RC4_40_MD5,					128,	40 },
{ SSL3_TXT_ADH_RC4_128_MD5,					128,	128 },
{ SSL3_TXT_ADH_DES_40_CBC_SHA,				40,		40 },
{ SSL3_TXT_ADH_DES_64_CBC_SHA,				56,		56 },
//{ SSL3_TXT_ADH_DES_196_CBC_SHA,				168,	168 },
{ SSL3_TXT_FZA_DMS_NULL_SHA,				0,		0 },
{ SSL3_TXT_FZA_DMS_FZA_SHA,					96,		96 },
{ SSL3_TXT_FZA_DMS_RC4_SHA,					96,		96 },

/* this must always be last */
{ 0, 0 }
};
static PIDB *pDBCiphers;

/* ---
BIO methods, hook in lower level IO object
--- */
static BIO_METHOD methods_PIIO_SSLeay=
	{
	BIO_TYPE_FILTER,			/* is this right? */
	"PIIOObject",
	PIIO_SSLeay_write,
	PIIO_SSLeay_read,
	PIIO_SSLeay_puts,
	NULL,	/* PIIO_SSLeay_gets, */
	PIIO_SSLeay_ctrl,
	PIIO_SSLeay_new,
	PIIO_SSLeay_free,
	};

/* ---
	Name of parameters
--- */
#define KEY_CONF_IOOBJECT				"IOObject"
#define KEY_CONF_VERSION				"Version"
#define KEY_CONF_VERIFY					"Verify"
#define KEY_CONF_VERIFYDEPTH			"VerifyDepth"
#define KEY_CONF_PRIVATEKEYFILE			"PrivateKeyFile"
#define KEY_CONF_CACERTIFICATEPATH		"CACertificatePath"
#define KEY_CONF_CACERTIFICATEFILE		"CACertificateFile"
#define KEY_CONF_CERTIFICATEFILE		"CertificateFile"
#define KEY_CONF_TYPE					"Type"
#define KEY_CONF_FLAG					"Flag"
#define KEY_CONF_DEBUGFILE				"DebugFile"
#define KEY_CONF_CIPHERLIST				"CipherList"
#define KEY_CONF_RANDOMSEEDFILE			"RandomSeedFile"
#define KEY_CONF_RANDOMSEEDSIZE			"RandomSeedSize"

#define VALUE_PASSIVE					"Passive"
#define VALUE_ACTIVE					"Active"
#define VALUE_BUGS						"Bugs"
#define VALUE_HACK						"Hack"
#define VALUE_DISABLE					"Disable"

/*____________________________________________________________________________*\
 *
 Description:
	Global data, Oh my Gaaaawd. Well this just about blows off
any chance of multiple SSL contexts operating without side effects...,
but I don't see any other way to do this given the nature of the SSL
interface.
\*____________________________________________________________________________*/
int verify_depth=0;
int verify_error=X509_V_OK;
static FILE *g_pStdioErr = NULL;
static BIO *g_pDbgErr = NULL;

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
#if ( PISSL_DEBUG )
//#if 0
void fnSSLInfo( SSL *s,int where, int ret)
{
	char *str;
	int w;

	w=where& ~SSL_ST_MASK;

	if (w & SSL_ST_CONNECT) str="SSL_connect";
	else if (w & SSL_ST_ACCEPT) str="SSL_accept";
	else str="undefined";

	if (where & SSL_CB_LOOP)
		{
		fprintf(stderr,"%s:%s\n",str,SSL_state_string_long(s));
		}
	else if (where & SSL_CB_ALERT)
		{
		str=(where & SSL_CB_READ)?"read":"write";
		fprintf(stderr,"SSL3 alert %s:%s:%s\n",
			str,
			SSL_alert_type_string_long(ret),
			SSL_alert_desc_string_long(ret));
		}
	else if (where & SSL_CB_EXIT)
		{
		if (ret == 0)
			fprintf(stderr,"%s:failed in %s\n",
				str,SSL_state_string_long(s));
		else if (ret < 0)
			{
			fprintf(stderr,"%s:error in %s\n",
				str,SSL_state_string_long(s));
			}
		}
} 
//#endif
#endif	/* DEBUG */

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	This function is the main reason that we have to mess around with
	those global variables - it offers no possibility to pass the 
	global debugging BIO * in as an argument, not that I can see anyway.
\*____________________________________________________________________________*/
int fnVerify( int ok, X509_STORE_CTX *ctx )
	{
	char buf[256];
	X509 *err_cert;
	int err,depth;

	err_cert=X509_STORE_CTX_get_current_cert(ctx);
	err=X509_STORE_CTX_get_error(ctx);
	depth=X509_STORE_CTX_get_error_depth(ctx);

	X509_NAME_oneline(X509_get_subject_name(err_cert),buf,256);
	BIO_printf(g_pDbgErr,"depth=%d %s\n",depth,buf);
	if (!ok)
		{
		BIO_printf(g_pDbgErr,"verify error:num=%d:%s\n",err,
			X509_verify_cert_error_string(err));
		if (verify_depth >= depth)
			{
			ok=1;
			verify_error=X509_V_OK;
			}
		else
			{
			ok=0;
			verify_error=X509_V_ERR_CERT_CHAIN_TOO_LONG;
			}
		}
	switch (ctx->error)
		{
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
		X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert),buf,256);
		BIO_printf(g_pDbgErr,"issuer= %s\n",buf);
		break;
	case X509_V_ERR_CERT_NOT_YET_VALID:
	case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
		BIO_printf(g_pDbgErr,"notBefore=");
		ASN1_UTCTIME_print(g_pDbgErr,X509_get_notBefore(ctx->current_cert));
		BIO_printf(g_pDbgErr,"\n");
		break;
	case X509_V_ERR_CERT_HAS_EXPIRED:
	case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
		BIO_printf(g_pDbgErr,"notAfter=");
		ASN1_UTCTIME_print(g_pDbgErr,X509_get_notAfter(ctx->current_cert));
		BIO_printf(g_pDbgErr,"\n");
		break;
		}
	if ( ctx->error != X509_V_OK )
	    {
	    ok = 0; 
	    };
	BIO_printf(g_pDbgErr,"verify return:%d\n",ok);
	return(ok);
	}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static RSA *fnTmpRSA(SSL *s, int is_export, int keylength)
{
	static RSA *rsa_tmp=NULL;
	PISSL *pSSL = (PISSL *)CRYPTO_get_ex_data(&s->ex_data, 0 );

	BIO *pDbgErr = pSSL->pConf->pBIODebug;

	if (rsa_tmp == NULL)
		{
		if (!s_quiet)
			{
			BIO_printf(pDbgErr,"Generating temp (%d bit) RSA key...", keylength);
			BIO_flush(pDbgErr);
			}
#ifndef NO_RSA
		rsa_tmp=RSA_generate_key(keylength,RSA_F4,NULL,NULL);
#endif
		if (!s_quiet)
			{
			BIO_printf(pDbgErr,"\n");
			BIO_flush(pDbgErr);
			}
		}
	return(rsa_tmp);
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
long fnBioDump( BIO *bio, int cmd, char *argp, int argi, long argl, long ret)
{
	PISSL *pSSL;
	BIO *out;

	pSSL = (PISSL *)BIO_get_callback_arg(bio);
	out=pSSL->pConf->pBIODebug;
	if (out == NULL) return(ret);

	if (cmd == (BIO_CB_READ|BIO_CB_RETURN))
		{
		BIO_printf(out,"read from %08X [%08lX] (%d bytes => %ld (0x%X))\n",
			bio,argp,argi,ret,ret);
		BIO_dump(out,argp,(int)ret);
		return(ret);
		}
	else if (cmd == (BIO_CB_WRITE|BIO_CB_RETURN))
		{
		BIO_printf(out,"write to %08X [%08lX] (%d bytes => %ld (0x%X))\n",
			bio,argp,argi,ret,ret);
		BIO_dump(out,argp,(int)ret);
		}
	return(ret);
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void fnSSLInfo( SSL *s, int where, int ret)
{
	char *str;
	int w;

	w=where& ~SSL_ST_MASK;

	if (w & SSL_ST_CONNECT) str="SSL_connect";
	else if (w & SSL_ST_ACCEPT) str="SSL_accept";
	else str="undefined";

	if (where & SSL_CB_LOOP)
		{
		BIO_printf(g_pDbgErr,"%s:%s\n",str,SSL_state_string_long(s));
		}
	else if (where & SSL_CB_ALERT)
		{
		str=(where & SSL_CB_READ)?"read":"write";
		BIO_printf(g_pDbgErr,"SSL3 alert %s:%s:%s\n",
			str,
			SSL_alert_type_string_long(ret),
			SSL_alert_desc_string_long(ret));
		}
	else if (where & SSL_CB_EXIT)
		{
		if (ret == 0)
			BIO_printf(g_pDbgErr,"%s:failed in %s\n",
				str,SSL_state_string_long(s));
		else if (ret < 0)
			{
			BIO_printf(g_pDbgErr,"%s:error in %s\n",
				str,SSL_state_string_long(s));
			}
		}
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int PIIO_SSLeay_write( BIO *h, char *pBuf, int iLen )
{
	PIObject *pIO;
	int iSent;
	int iRet;

	/* --- get the lower level IO object --- */
	pIO = (PIObject *)CRYPTO_get_ex_data(&(h->ex_data), 0 );
	iRet = PIIO_out( pIO, pBuf, iLen, &iSent, 0, 0 );
	switch( iRet )
		{
		case PIAPI_COMPLETED:
			return iSent;

		case PIAPI_ERROR:
		default:
			return -1;	
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int PIIO_SSLeay_read(BIO *h, char *pBuf, int iLen )
{
	PIObject *pIO;
	int iRead;
	int iRet;

	/* --- get the lower level IO object --- */
	pIO = (PIObject *)CRYPTO_get_ex_data(&(h->ex_data), 0 );
	iRet = PIIO_in( pIO, pBuf, iLen, &iRead, 0, 0 );
	switch( iRet )
		{
		case PIAPI_COMPLETED:
			return iRead;

		case PIAPI_ERROR:
		default:
			return -1;	
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int PIIO_SSLeay_puts( BIO *h, char *str )
{
	int iLen;

	assert( str );
	iLen = strlen( str );

	return PIIO_SSLeay_write( h, str, iLen );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static long PIIO_SSLeay_ctrl(BIO *h,int cmd,long arg1,char *arg2)
{
	/* --- is this right ? --- */
	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int PIIO_SSLeay_new(BIO *bi)
{
	bi->init=1;
	bi->num=0;
	bi->ptr=NULL;
	bi->flags=0;
	return(1);
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static int PIIO_SSLeay_free(BIO *bi)
{
	bi->init=0;
	bi->flags=0;
	return(1);
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int SSL_fnParameter( void *pData, const char *pVar, const char *pVal,
	const char *pWhere )
{
	PIObject *pObject = (PIObject *)pData;
	PISSL *pSSL = (PISSL *)PIObject_getUserData( pObject );
	SSLConf *pConf = pSSL->pConf;
	assert( pVar );
	assert( pVal );

	if ( !PIUtil_stricmp( pVar, KEY_CONF_VERSION ) )
		{
		int iTmp = atoi( pVal );
		if ( iTmp!=2 && iTmp!=3 && iTmp!=23 && iTmp!=10 )
			{
			PILog_addMessage(
				PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ),
				PILOG_ERROR, "%sSSL: Bad SSL version %d. Accepted \
version numbers are 2, 3 or 23 or 10 for TLS1.0.", pWhere, iTmp );
			return 0;
			};
		pConf->iVersion = iTmp;
		}
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_VERIFY ) )
		{
		int iTmp = atoi( pVal );
		switch( iTmp )
			{
			case 0:
				iTmp = SSL_VERIFY_NONE; break;

			case 1:
				iTmp = SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE; break;

			case 2:
				iTmp = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT; break;

			default:
				PILog_addMessage(
					PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ),
					PILOG_ERROR, "%sSSL: Bad SSL verification code %d. \
Accepted verification codes are 0, 1, 2 or 3.", pWhere, iTmp );
				return 0;
			};
		pConf->iVerify = iTmp;
		}
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_VERIFYDEPTH ) )
		{ pConf->iVerifyDepth = atoi( pVal ); }
	else if ( !PIUtil_stricmp( KEY_CONF_TYPE, pVar ) )
		{ pConf->iServer = PIUtil_stricmp( pVal, VALUE_PASSIVE ) ? 0 : 1; }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_PRIVATEKEYFILE ) )
		{ pConf->pKeyF = PIUtil_strdup( pVal );	}
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_CACERTIFICATEPATH ) )
		{ pConf->pCACertP = PIUtil_strdup( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_CACERTIFICATEFILE ) )
		{ pConf->pCACertF = PIUtil_strdup( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_CERTIFICATEFILE ) )
		{ pConf->pCertF = PIUtil_strdup( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_DEBUGFILE ) )
		{ pConf->pDebugFile = PIUtil_strdup( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_CIPHERLIST ) )
		{ pConf->pCipherList = PIUtil_strdup( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_RANDOMSEEDFILE ) )
		{ pConf->pRandomSeedFile = PIUtil_strdup( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_RANDOMSEEDSIZE ) )
		{ pConf->iRandomSize = atoi( pVal ); }
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_FLAG ) )
		{
		if ( !PIUtil_stricmp( pVal, VALUE_BUGS ) )
			{ pConf->iBugs = 1; }
		else if ( !PIUtil_stricmp( pVal, VALUE_HACK ) )
			{ pConf->iHack = 1; }
		else if ( !PIUtil_stricmp( pVal, VALUE_DISABLE ) )
			{ pConf->iDisable = 1; }
		else
			{
			PILog_addMessage(
				PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ),
				PILOG_ERROR, "%sSSL: Unknown flag '%s'.", pWhere, pVal );
			return 0;
			};
		}
	else if ( !PIUtil_stricmp( pVar, KEY_CONF_IOOBJECT ) )
		{
		/* 
		Load the prototype for the transport IO object. The function
		PIObject_loadFromLine() will log the appropriate error messages
		if it fails
		*/
		pSSL->pIOObject = PIObject_loadFromLine( 
			PIObject_getDB( pObject ),
			PIObject_getConfigurationDB( pObject ),
			pVal );
		if ( !pSSL->pIOObject )
			{ return 0; };
		}
	else
		{
		/* ---
		Configuration error
		--- */
		PILog_addMessage(
			PIObject_getDB( pObject ),
			PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "%sSSL: Unknown directive '%s'.", pWhere, pVar );
		return 0;
		};
	return 1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return non-zero on success, zero  on failure.
\*____________________________________________________________________________*/
int SSL_init( PIObject *pObject, int iArgc, const char *ppArgv[] )
{
	PISSL *pPrototype;
	SSLConf *pConf;
	int iRet;
	SSL_METHOD *pSSLMethod = 0;
	SSL *pSSL;
	X509 *pX509;
	PIPLATFORM_FD tFD;
	void *pBuf;
	static const char rnd_seed[] = "string to make the random number generator think it has entropy";

	pPrototype = PIObject_getUserData( pObject );
	pConf = pPrototype->pConf;
	iRet = PIObject_readParameters( pObject, iArgc, ppArgv,
		SSL_fnParameter, pObject );

	if ( !iRet )
		{ 
		goto init_error;
		};

	/* ---
	Initialize the random number generator
	or OAEP may fail
	--- */
	if ( pConf->pRandomSeedFile && !pConf->iDisable )
		{
		/* --- get file info --- */
		PIFInfo *pFInfo = PIFInfo_new( pConf->pRandomSeedFile );

		if ( !pFInfo || !PIFInfo_exists( pFInfo ) || PIFInfo_isDirectory( pFInfo )
			|| !PIFInfo_isReadable( pFInfo))
			{
			PILog_addMessage(
				PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ),
				PILOG_ERROR, "SSL: random seed file '%s' \
doesn't exist or isn't readable.",
				pConf->pRandomSeedFile );
			goto init_error;
			};

		/* --- open file handle --- */
		if ((tFD = PIFile_open( PIFInfo_getPath( pFInfo ), "r" )) == PIPLATFORM_FD_INVALID)
			{
			/* --- handle error --- */
			int iErrorCode = PIPlatform_getLastError();
			PILog_addMessage(
				PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ),
				PILOG_ERROR, "SSL: Error on open random seed file '%s'. \
The system error code is %d.",
				pConf->pRandomSeedFile, iErrorCode );
			goto init_error;
			};

		pBuf = PIUtil_malloc( pConf->iRandomSize );
		if ( !pBuf )
			{
			/* --- handle error --- */
			int iErrorCode = PIPlatform_getLastError();
			PILog_addMessage(
				PIObject_getDB( pObject ),
				PIObject_getConfigurationDB( pObject ),
				PILOG_ERROR, "SSL: Error on allocate random seed buffer. \
The system error code is %d.", iErrorCode );
			goto init_error;
			};

		PIFile_read( tFD, pConf->iRandomSize, pBuf );
		PIFile_close( tFD );

		RAND_seed( pBuf, pConf->iRandomSize );
		PIUtil_free( pBuf );
		}
	else
		{
		RAND_seed(rnd_seed, sizeof rnd_seed);
		};

	/*
	** Open a debug file ?
	*/
	if ( pConf->pDebugFile )
		{
		if ( g_pDbgErr )
			{
			/*
			** Debug file is already open
			*/
			pConf->pBIODebug = g_pDbgErr;
			g_pDbgErr = pConf->pBIODebug;
			}
		else if ( !strcmp( pConf->pDebugFile, "STDERR" ) )
			{
			pConf->pBIODebug = BIO_new_fp(stderr,BIO_NOCLOSE);
			g_pDbgErr = pConf->pBIODebug;
			}
		else if ( !strcmp( pConf->pDebugFile, "STDOUT" ) )
			{
			pConf->pBIODebug = BIO_new_fp(stdout,BIO_NOCLOSE);
			g_pDbgErr = pConf->pBIODebug;
			}
		else
			{
			pConf->pStdioDebug = fopen( pConf->pDebugFile, "w" );
			if ( !pConf->pStdioDebug )
				{
				PILog_addMessage(
					PIObject_getDB( pObject ),
					PIObject_getConfigurationDB( pObject ),
					PILOG_ERROR, "SSL: Failed to open debug file '%s'.",
					pConf->pDebugFile );
				goto init_error;
				};
			pConf->pBIODebug = BIO_new_fp( pConf->pStdioDebug, BIO_NOCLOSE );
			g_pStdioErr = pConf->pStdioDebug;
			g_pDbgErr = pConf->pBIODebug;
			};
		}

	/* ---
	Check values were set
	--- */
	if ( !pPrototype->pIOObject )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'IOObject' not defined." );
		goto init_error;
		};

	if ( !pConf->iServer && !pConf->pKeyF )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ),
			PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'PrivateKeyFile' must be defined \
for server connection." );
		goto init_error;
		};

	if ( !pConf->iServer && !pConf->pCertF )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ),
			PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'CertificateFile' must be defined \
for server connection." );
		goto init_error;
		};

	/*
	** Funny little piece of code
	*/
	switch( pConf->iVersion * ( 1 + pConf->iServer ) )
		{
		case 10: pSSLMethod = TLSv1_client_method(); break;
		case 23: pSSLMethod = SSLv23_client_method(); break;
		case 2: pSSLMethod = SSLv2_client_method(); break;
		case 3: pSSLMethod = SSLv3_client_method(); break;

		case 20: pSSLMethod = TLSv1_server_method(); break;
		case 46: pSSLMethod = SSLv23_server_method(); break;
		case 4: pSSLMethod = SSLv2_server_method(); break;
		case 6: pSSLMethod = SSLv3_server_method(); break;

		default:
			assert( 0 );
		};
	pConf->pCtx = SSL_CTX_new( pSSLMethod );
	if ( !pConf->pCtx )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'SSL_CTX_new()' failed." );
		INIT_ERR;
		goto init_error;
		};
 
#if ( PISSL_DEBUG )
	SSL_CTX_set_info_callback( pConf->pCtx, fnSSLInfo );
#endif

	/*
	** Set list of allowed ciphers
	*/
	if (	pConf->pCipherList &&
			!SSL_CTX_set_cipher_list( pConf->pCtx, pConf->pCipherList ) )
		{
		/*
		** Cipher list was specified, but failed to set it
		*/
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'SSL_set_cipher_list()' failed \
for cipher list '%s'.", pConf->pCipherList );
		INIT_ERR;
		ERR_print_errors_fp(stderr); 
		if ( g_pStdioErr ) { ERR_print_errors_fp(g_pStdioErr); }; 
		goto init_error;
		};

	/*
	** What are these for??
	*/
	if ( pConf->iBugs )
		{ SSL_CTX_set_options( pConf->pCtx, SSL_OP_ALL); };
	if ( pConf->iHack )
		{
		SSL_CTX_set_options( pConf->pCtx,
			SSL_OP_NETSCAPE_DEMO_CIPHER_CHANGE_BUG);
//		SSL_CTX_set_options( pConf->pCtx, SSL_OP_NON_EXPORT_FIRST);
		};

	if ( (!SSL_CTX_load_verify_locations( pConf->pCtx, pConf->pCACertF,
			pConf->pCACertP ) ) ||
		 (!SSL_CTX_set_default_verify_paths( pConf->pCtx ) ) )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'SSL_CTX_load_verify_locations()' failed." );
		INIT_ERR;
		goto init_error;
		}
  
	/*
	** Set SSL verification callback
	*/
	SSL_CTX_set_verify( pConf->pCtx, pConf->iVerify, fnVerify );

	if ( SSL_CTX_use_certificate_file(pConf->pCtx, pConf->pCertF,
		SSL_FILETYPE_PEM) ==-1 )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'SSL_CTX_use_certificate_file()' failed." );
		INIT_ERR;
		goto init_error;
		};

	if ( SSL_CTX_use_PrivateKey_file(pConf->pCtx, pConf->pKeyF,
		SSL_FILETYPE_PEM) ==-1 )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: 'SSL_CTX_use_PrivateKey_file()' failed." );
		INIT_ERR;
		goto init_error;
		};

	SSL_CTX_set_tmp_rsa_callback( pConf->pCtx, fnTmpRSA );

	/*
	** If we are using DSA, we can copy the parameters from
	** the private key
	*/
	pSSL=SSL_new( pConf->pCtx );
	assert( pSSL );
	CRYPTO_set_ex_data(&pSSL->ex_data, 0, (char *)pPrototype );
	pX509=SSL_get_certificate( pSSL );
	if ( pX509 )
		{
		EVP_PKEY_copy_parameters(
			X509_get_pubkey( pX509 ), SSL_get_privatekey( pSSL ) );
		};
	SSL_free( pSSL );

	/*
	** Now we know that a key and cert have been set against
	** the SSL context
	*/
	if ( !SSL_CTX_check_private_key( pConf->pCtx ) )
		{
		PILog_addMessage(
			PIObject_getDB( pObject ), PIObject_getConfigurationDB( pObject ),
			PILOG_ERROR, "SSL: Private key does not match the \
certificate public key." );
		INIT_ERR;
		goto init_error;
		}

	return iRet;

init_error:
	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SSL_onClassLoad( PIClass_LoadAction eLoad, int i,
	const char *a[] )
{
	if ( eLoad==STARTUP )
		{
		int i;

		/*
		** Once off initialization for class
		*/
		assert( !g_pDbgErr );

		/*
		** Strings and algorithms
		*/
		SSL_load_error_strings();
		SSLeay_add_ssl_algorithms();

		/*
		** Give the SSLeay message on startup
		*/
		fprintf(stdout, "This product includes software developed by the OpenSSL Project \
for use in the OpenSSL Toolkit (http://www.openssl.org/).\n" );

		/*
		** Add cipher information
		*/
		pDBCiphers = PIDB_new( 0, "Ciphers" );
		for( i=0; aCiphers[i].pName; i++ )
			{
			PIDB_add( pDBCiphers, PIDBTYPE_OPAQUE, aCiphers[i].pName,
				(void *)&(aCiphers[i]), 0 );
			};
		}
	else if ( eLoad==SHUTDOWN )
		{
		PIDB_delete( pDBCiphers );

		if ( g_pDbgErr )
			{
			BIO_flush( g_pDbgErr );
			BIO_free( g_pDbgErr );
			g_pDbgErr = NULL;
			};
		if ( g_pStdioErr )
			{
			fclose( g_pStdioErr );
			g_pStdioErr = NULL;
			};
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SSL_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	SSLConf *pConf;
	PISSL *pPrototype;

	pConf = PIUtil_malloc( sizeof( SSLConf ) );
	memset( pConf, 0, sizeof( SSLConf ) );
	pConf->iVersion = 23;
	pConf->iVerify = SSL_VERIFY_NONE;
	pConf->iVerifyDepth = 8;
	pConf->iBugs = 0;
	pConf->iHack = 0;
	pConf->iDisable = 0;
	
	/* ---
	Initialize the IOfilterConf structure
	--- */
	pConf->iServer = 1;

	pPrototype = PIUtil_malloc( sizeof( PISSL ) );
	memset( pPrototype, 0, sizeof( PISSL ) );
	pPrototype->pConf = pConf;
	pPrototype->iIsPrototype = 1;
	PIObject_setUserData( pObj, pPrototype );
	if ( !SSL_init( pObj, iArgc, ppArgv ) )
		{
		return PIAPI_ERROR;
		};

	/*
	** Ugly global variables! at least try to warn of possible
	** collision
	*/
	assert( verify_depth==0 );
	verify_depth = pConf->iVerifyDepth;

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SSL_copyConstructor( PIObject *pObj, int iArgc,
	const char *ppArgv[] )
{
	BIO *pBio;
	PISSL *pPrototypeSSL;
	PISSL *pChildSSL;
	int iTmp;
	char *pCipher;
	CipherInformation *pCipherInfo;

	/* ---
	Copy constructor 
	--- */
	pPrototypeSSL = PIObject_getUserData( pObj );
	pChildSSL = PIUtil_malloc( sizeof( PISSL ) );
	memset( pChildSSL, 0, sizeof( PISSL ) );
	PIObject_setUserData( pObj, pChildSSL );
	pChildSSL->iError = 1;		/* set error flag */
	pChildSSL->pConf = pPrototypeSSL->pConf;
	pChildSSL->iIsPrototype = 0;
	pChildSSL->pIOObject = PIObject_copy(
		pPrototypeSSL->pIOObject,
		iArgc,
		ppArgv
		);

	/* ---
	Successful?
	--- */
	if ( !pChildSSL->pIOObject )
		{
		goto setup_error;
		};

	/* ---
	Save this objects DB and then set this DB to be that of the
	contained IO object, this allows the callee to access parameters
	in the lower level IO object chain
	--- */
	pChildSSL->pSaveDB = PIObject_getDB( pObj );
	PIObject_setDB( pObj, PIObject_getDB( pChildSSL->pIOObject ) );

	// SSL is switched off
	if ( pChildSSL->pConf->iDisable ) return PIAPI_COMPLETED;

	/* ---
	Setup the SSL portion of this request
	--- */
	pBio = BIO_new( &methods_PIIO_SSLeay );
	pChildSSL->pBio = pBio;
	CRYPTO_set_ex_data( &(pBio->ex_data), 0, (char *)pChildSSL->pIOObject ); 
	assert( pChildSSL->pConf->pCtx );
	pChildSSL->pSSL = SSL_new( pChildSSL->pConf->pCtx );
	if ( !pChildSSL->pSSL )
		{
		RUNTIME_ERR( "SSL_new() failed." );
		goto setup_error;
		};

	CRYPTO_set_ex_data( &(pChildSSL->pSSL)->ex_data, 0, (char *)pChildSSL );

	/*
	** Set debug callback ?
	*/
	if ( pChildSSL->pConf->pDebugFile )
		{
		BIO_set_callback( pBio, fnBioDump );
		BIO_set_callback_arg( pBio, pChildSSL );
		SSL_CTX_set_info_callback( pChildSSL->pConf->pCtx, fnSSLInfo );
		};

	SSL_set_bio( pChildSSL->pSSL, pBio, pBio );
	if ( pChildSSL->pConf->iServer )
		{
		/*
		** Server connection, do SSL_accept() 
		*/
		SSL_set_verify( pChildSSL->pSSL, pChildSSL->pConf->iVerify, NULL );
		iTmp = SSL_accept( pChildSSL->pSSL );

		if ( iTmp==-1 )
			{
			RUNTIME_ERR( "SSL_accept() failed." );
			goto setup_error;
			};
		}
	else
		{
		/*
		** Client connection, do SSL_connect()
		*/
		iTmp = SSL_connect( pChildSSL->pSSL );
		if ( iTmp==-1 )
			{
			RUNTIME_ERR( "SSL_connect() failed." );
			goto setup_error;
			};
		};
	
	/*
	** Set SSL DB values
	*/
	pCipher = SSL_get_cipher( pChildSSL->pSSL );
	pCipherInfo = PIDB_lookup( pDBCiphers, PIDBTYPE_OPAQUE,
		pCipher, 0 );
	if ( pCipherInfo )
		{
		enum { BUF_SIZE=63 };
		char szBuf[ BUF_SIZE + 1 ];

		/*
		** Set keysize and cipher name, the HTTP logic components
		** can generate the correct environment variables given these
		*/
		PIDB_add( PIObject_getDB( pObj ), PIDBTYPE_STRING, "CipherName",
			(void *)pCipherInfo->pName, 0 );
		sprintf( szBuf, "%d", pCipherInfo->iKeySize );
		PIDB_add( PIObject_getDB( pObj ), PIDBTYPE_STRING, "KeySize", szBuf,
			0 );
		sprintf( szBuf, "%d", pCipherInfo->iSecretKeySize );
		PIDB_add( PIObject_getDB( pObj ), PIDBTYPE_STRING, "SecretKeySize",
			szBuf, 0 );
		}
	else
		{
		fprintf(stderr,"SSL: Unknown cipher, '%s'\n", pCipher ? pCipher : "" );
		goto setup_error;
		};

	/*
	** Clear error flag
	*/
	pChildSSL->iError = 0;

setup_error:
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:   	SSL_out
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SSL_out( PIObject *pObj, void *pData, int iLen,
		int *piSent, int iArgc, const char *ppArgv[] )
{
	PISSL *pPISSL;
	int iRet;

	pPISSL = (PISSL *)PIObject_getUserData( pObj );

	assert( pPISSL );
	if ( pPISSL->pConf->iDisable )
		{
		iRet = PIIO_out( pPISSL->pIOObject, pData, iLen, piSent, 0, 0 );
		return iRet;
		}
	else
		{
		if ( pPISSL->iError )
			{ return PIAPI_ERROR; };

		assert( pPISSL->pSSL );
		iRet = SSL_write( pPISSL->pSSL, (char *)pData, iLen );
		};

	if ( iRet>=0 )
		{
		*piSent = iRet;
		return PIAPI_COMPLETED;
		};
	
	/* else */
	return PIAPI_ERROR;
}

/*____________________________________________________________________________*\
 *
 Function:   	SSL_in
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SSL_in( PIObject *pObj, void *pData, int iLen,
		int *piReceived, int iArgc, const char *ppArgv[] )
{
	PISSL *pPISSL;
	int iRet;

	pPISSL = (PISSL *)PIObject_getUserData( pObj );

 	assert( pPISSL );
	if ( pPISSL->pConf->iDisable )
		{
		iRet = PIIO_in( pPISSL->pIOObject, pData, iLen, piReceived, 0, 0 );
		return iRet;
		}
	else
		{
		if ( pPISSL->iError )
			{ return PIAPI_ERROR; };

		assert( pPISSL->pSSL );
		iRet = SSL_read( pPISSL->pSSL, (char *)pData, iLen );
		};

	if ( iRet>=0 )
		{
		*piReceived = iRet;
		return PIAPI_COMPLETED;
		};
	
	/* else */
	return PIAPI_ERROR;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SSL_destructor( PIObject *pObj, int i, const char *a[] )
{
	PISSL *pPISSL = (PISSL *)PIObject_getUserData( pObj );

	if ( pPISSL->iIsPrototype )
		{
		PIUtil_free( pPISSL->pConf->pCACertP );
		PIUtil_free( pPISSL->pConf->pCACertF );
		PIUtil_free( pPISSL->pConf->pKeyF );
		PIUtil_free( pPISSL->pConf->pCertF );
		PIUtil_free( pPISSL->pConf->pDebugFile );
		if ( pPISSL->pConf->pBIODebug )
			{
			fflush(pPISSL->pConf->pBIODebug->ptr);
//			BIO_flush( pPISSL->pConf->pBIODebug );
/*
** Can't do this because the file object is global
*/
/*			fclose( pPISSL->pConf->pStdioDebug ); */
			fflush( pPISSL->pConf->pStdioDebug );
			};
		SSL_CTX_free( pPISSL->pConf->pCtx );
		PIUtil_free( pPISSL->pConf->pCipherList );
		PIUtil_free( pPISSL->pConf );
		}
	else
		{
		if ( pPISSL->pBio )
			{ BIO_flush( pPISSL->pBio ); };

		if ( pPISSL->pSSL )
			{ SSL_free( pPISSL->pSSL ); };

		/* ---
		Restore DB
		--- */
		if ( pPISSL->pSaveDB )
			{ PIObject_setDB( pObj, pPISSL->pSaveDB ); };
		};

	if ( pPISSL->pIOObject )
		{ PIObject_delete( pPISSL->pIOObject, 0, 0 ); };

	PIUtil_free( pPISSL );

	return PIAPI_COMPLETED;
}

/* ---
	Class configuration specification 
--- */
#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name SSLClass
		Type IO
		Library SSL
		OnClassLoad SSL_onClassLoad
		Constructor SSL_constructor
		CopyConstructor SSL_copyConstructor
		Destructor SSL_destructor
		Out SSL_out
		In SSL_in
	</Class>

	<Object>
		Name SSL
		Class SSLClass
	</Object>

/*___+++CNF_END+++___*/
#endif
