/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Server/IOSrvBse.cpp,v $
 * $Date: 2001/04/07 13:30:22 $
 *
 Description:
	Base class and code for IO servers.
\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>
#include <signal.h>

#include "IOSrvBse.h"
#include "Pi2API.h"

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
#define MODULE_NAME				"IOServer"
#define KEY_CONF_IOOBJECT		"IOObject"
#define KEY_CONF_LOGICOBJECT	"LogicObject"
#define CONFIG_ERR(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_ERROR, (msg) ); }
#define CONFIG_WARN(this, msg)   \
	{ PILog_addMessage( PIObject_getDB(this), \
	PIObject_getConfigurationDB(this), PILOG_WARNING, (msg) ); }

/*
#define D { cerr << (int)PIThread_getSystemHandle(PIThread_getCurrent()) \
		<< ": " << __FILE__ << ", " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	private
 Description:
\*____________________________________________________________________________*/
int IOServerBase::DoDispatchConnection( PIObject *pIOObject )
{
	int iRet = DispatchConnection( pIOObject );
	if ( iRet!=PIAPI_COMPLETED )
		{
		CONFIG_ERR( pObject, "IOServer: DoDispatchConnection() failed.");
		};
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	private
 Description:
\*____________________________________________________________________________*/
int IOServerBase::DispatchConnection( PIObject *pIOObject )
{
	int iRet = PILogic_execute( pLogic, 1, (const char **)&pIOObject );
	PIObject_delete( pIOObject, 0, 0 );
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
IOServerBase::IOServerBase( const PIObject *pTheObject )
:	pObject( pTheObject ),
	pLogic( 0 ),
	pIO( 0 ),
	bOK( 0 ),
	pTree( 0 )
{
	PIConfig *pConfigDB = PIObject_getConfigurationDB( pObject );
	pTree = PIDB_new( PIObject_getDB( pObject ), MODULE_NAME );
	if ( !pTree )
		{
		CONFIG_ERR( pObject, "IOServer: PIDB_new() failed");
		return;
		};

	pIO = PIObject_loadEx( pObject, KEY_CONF_IOOBJECT );

	const char *pLogicName = PIConfig_lookupValue( pConfigDB,
		KEY_CONF_LOGICOBJECT, 0, 0 );
	if ( !pLogicName )
		{
		CONFIG_ERR( pObject, "IOServer: Missing definition for 'LogicObject'");
		return;
		};

	pLogic = PIObject_load(
		pTree,
		0,
		pLogicName,
		0,
		0 );
	if ( pIO && pLogic )
		{
		/* --- do some final details --- */
#if POSIX
		/* ---
		Ignore SIG_PIPE on Unix systems. This signal is send whenever a 
	 	pipe or socket is closed while there is still data to be read
		(data is discarded).
		--- */
		bOK = ( signal( SIGPIPE, SIG_IGN )!=SIG_ERR );
#elif WIN32
		bOK = 1;
#endif
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
IOServerBase::~IOServerBase()
{
	PIObject_delete( pLogic, 0, 0 );
	PIObject_delete( pIO, 0, 0 );
	PIDB_delete( pTree );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int IOServerBase::Execute()
{
	int iRet = PIAPI_COMPLETED;
	for(; iRet!=PIAPI_ABORT;)
		{
		iRet = Accept();
		};
	return iRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int IOServerBase::Accept()
{
	/* ---
	Accept a new child IO object
	--- */
	PIObject *pChildIO = PIObject_copy( pIO,
		0, 0 );
	if ( !pChildIO )
		{
		/* --- consider this a critical error --- */
		CONFIG_ERR( pObject, "IOServerBase: PIObject_copy() failed for \
IO Object");
		return PIAPI_ABORT;
		};

	return DoDispatchConnection( pChildIO );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int IOServerBase_execute( PIObject *pObj, int, const char *[] )
{
	if ( !pObj ) return PIAPI_ERROR;
	return ((IOServerBase *)PIObject_getUserData(pObj))->Execute();
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int IOServerBase_destructor( PIObject *pObj, int, const char *[] )
{
	delete (IOServerBase *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}

