/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
	mentioning features must contain the following
	acknowledgment:
	"This product includes software developed by John Roy
	and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/org/pi3/jsp/core/Pi3PageContext.java,v $
 * $Date: 2001/08/19 18:40:44 $
 *
 Description:
	Implementation of jsp API PageContext interface.

\*____________________________________________________________________________*/
 
package org.pi3.jsp.core;

import java.io.IOException;
import java.lang.IllegalArgumentException;
import java.lang.reflect.Array;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.RequestDispatcher;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import javax.servlet.jsp.JspWriter;
import javax.servlet.jsp.PageContext;

import org.pi3.servlet.core.Pi3HttpServletRequest;
import org.pi3.servlet.core.Pi3HttpSession;

/**
 * <p>
 * A PageContext instance provides access to all the namespaces associated with
 * a JSP page, provides access to several page attributes, as well as a layer above the
 * implementation details.
 * <p>
 * An instance of an implementation dependent subclass of this abstract base
 * class is created by a JSP implementation class at the begining of it's
 * <code> _jspService() </code> method via an implementation default 
 * <code> JspFactory </code>, as follows:
 *</p>
 *<p>
 *<p>
 *<code>
 *<pre>
 * public class foo implements Servlet {
 *
 * // ...
 *
 *public void _jspService(HttpServletRequest request,
 *			HttpServletResponse response)
 *       throws IOException, ServletException {
 *
 *    JspFactory  factory     = JspFactory.getDefaultFactory();
 *    PageContext pageContext = factory.getPageContext(
					this,
					request,
					response,
					null,  // errorPageURL
					false, // needsSession
					JspWriter.DEFAULT_BUFFER,
					true   // autoFlush
			        );
 *
 *    // initialize implicit variables for scripting env ...
 *
 *    HttpSession session = pageContext.getSession();
 *    JspWriter   out     = pageContext.getOut();
 *    Object      page    = this;
 *
 *    try {
 *        // body of translated JSP here ...
 *    } catch (Exception e) {
 *        out.clear();
 *        pageContext.handlePageException(e);
 *    } finally {
 *        out.close();
 *	  factory.releasePageContext(pageContext);
 *    }
 *}
 *</pre>
 * </code>
 * </p>
 * <p>
 * The <code> PageContext </code> class is an abstract class, designed to be
 * extended to provide implementation dependent implementations thereof, by
 * conformant JSP engine runtime environments. A PageContext instance is 
 * obtained by a JSP implementation class by calling the JspFactory.getPageContext() method, and is released by calling JspFactory.releasePageContext().
 * </p>
 * <p>
 * The PageContext provides a number of facilities to the page/component author and
 * page implementor, including:
 * <td>
 * <li>a single API to manage the various scoped namespaces
 * <li>a number of convenience API's to access various public objects
 * <li>a mechanism to obtain the JspWriter for output
 * <li>a mechanism to manage session usage by the page
 * <li>a mechanism to expose page directive attributes to the scripting environment
 * <li>mechanisms to forward or include the current request to other active components in the application
 * <li>a mechanism to handle errorpage exception processing
 * </td>
 * </p>
 *
 */

public class Pi3PageContext extends PageContext {

    private Servlet srl;
	private ServletRequest req;
	private ServletResponse res;
	private String errPage;
	private boolean nSess = true;
	private JspWriter out;
	private Hashtable attr = new Hashtable();

    /**
     * <p>
     * The initialize method is called to initialize an uninitialized PageContext
     * so that it may be used by a JSP Implementation class to service an
     * incoming request and response wihtin it's _jspService() method.
     * </p>
     * <p>
     * This method is typically called from JspFactory.getPageContext() in
     * order to initialize state.
     * </p>
     *
     * @param servlet The Servlet that is associated with this PageContext
     * @param request The currently pending request for this Servlet
     * @param response The currently pending response for this Servlet
     * @param errorPageURL The value of the errorpage attribute from the page directive or null
     * @param needsSession The value of the session attribute from the page directive
     * @param bufferSize The value of the buffer attribute from the page directive
     * @param autoFlush The value of the autoflush attribute from the page directive
     *
     * @throws IOException during creation of JspWriter
     * @throws IllegalStateException if out not correctly initialized
     */
 
    public void initialize(Servlet servlet, ServletRequest request, ServletResponse response, String errorPageURL, boolean needsSession, int bufferSize, boolean autoFlush)  throws IOException, IllegalStateException, IllegalArgumentException {
        srl = servlet;
		req = request;
		res = response;
		errPage = errorPageURL;
		nSess = needsSession;
		out =  new Pi3JspWriter(res.getOutputStream(), bufferSize, autoFlush);
	}

    /**
     * <p>
     * This method shall "reset" the internal state of a PageContext, releasing
     * all internal references, and preparing the PageContext for potential
     * reuse by a later invocation of initialize(). This method is typically
     * called from JspFactory.releasePageContext().
     * </p>
     *
     * <p> subclasses shall envelope this method </p>
     *
     */

    public void release() {
        srl = null;
		req = null;
		res = null;
		errPage = null;
		nSess = true;
		out = null;
		attr = null;
	}




    /**
     * register the name and object specified with page scope semantics
     * 
     * @throws NullPointerException if the name or object is null
     */

    public void setAttribute(String name, Object attribute) {
		attr.put(name, attribute);
	};




    /**
     * register the name and object specified with appropriate scope semantics
     * 
     * @param name the name of the attribute to set
     * @param o    the object to associate with the name
     * @param scope the scope with which to associate the name/object
     * 
     * @throws NullPointerException if the name or object is null
     * @throws IllegalArgumentException if the scope is invalid
     *
     */

    public void setAttribute(String name, Object o, int scope) {
		switch (scope) {
			case 1 /* PAGE_SCOPE */ :
				attr.put(name, o); break;

			case 2 /* REQUEST_SCOPE */ :
				req.setAttribute(name, o); break;

			case 3 /* SESSION_SCOPE */ :
				if (getSession() != null) getSession().putValue(name, o); break;

			case 4 /* APPLICATION_SCOPE */ :
				getServletContext().setAttribute(name, o); break;

			default: throw new IllegalArgumentException("Invalid scope.");
		};
	};




    /**
     * <p>return the object associated with the name in the page scope or null </p>
     *
     * @param name the name of the attribute to get
     * 
     * @throws NullPointerException if the name is null
     * @throws IllegalArgumentException if the scope is invalid
     */

    public Object getAttribute(String name) {
		return attr.get(name);
	};




    /**
     * <p>return the object associated with the name in the specifed scope or null </p>
     *
     * @param name the name of the attribute to set
     * @param scope the scope with which to associate the name/object
     * 
     * @throws NullPointerException if the name is null
     * @throws IllegalArgumentException if the scope is invalid
     */

    public Object getAttribute(String name, int scope) {
		switch (scope) {
			case 1 /* PAGE_SCOPE */ :
				return attr.get(name);

			case 2 /* REQUEST_SCOPE */ :
				return req.getAttribute(name);

			case 3 /* SESSION_SCOPE */ :
				return (getSession() != null) ?	getSession().getValue(name) : null;

			case 4 /* APPLICATION_SCOPE */ :
				return getServletContext().getAttribute(name);

			default: throw new IllegalArgumentException("Invalid scope.");
		};
	};




    /**
     * <p>
     * Searches for the named attribute in page, request, session (if valid),
     * and application scope(s) in order and returns the value associated or
     * null.
     * </p>
     *
     * @return the value associated or null
     */

    public Object findAttribute(String name) {
		Object o = attr.get(name);
		if (o != null) return o;
		
		o = req.getAttribute(name);
		if (o != null) return o;

		o = (getSession() != null) ? getSession().getValue(name) : null;
		if (o != null) return o;
		
		return getServletContext().getAttribute(name);
	};




    /**
     * remove the object reference associated with the specified name
     */

    public void removeAttribute(String name) {
		attr.remove(name);
	};




    /**
     * remove the object reference associated with the specified name
     */

    public void removeAttribute(String name, int scope) {
		switch (scope) {
			case 1 /* PAGE_SCOPE */ :
				attr.remove(name); break;

			case 2 /* REQUEST_SCOPE */ :
				((Pi3HttpServletRequest)req).removeAttribute(name); break;

			case 3 /* SESSION_SCOPE */ :
				if (getSession() != null) getSession().removeValue(name); break;

			case 4 /* APPLICATION_SCOPE */ :
				getServletContext().removeAttribute(name); break;

			default: throw new IllegalArgumentException("Invalid scope.");
		};
	};




    /**
     * @return the scope of the object associated with the name specified or 0
     */

    public int getAttributesScope(String name) {
		if (attr.get(name) != null) return PageContext.PAGE_SCOPE;
		if (req.getAttribute(name) != null) return PageContext.REQUEST_SCOPE;
		if ((getSession() != null) && (getSession().getValue(name) != null)) return PageContext.SESSION_SCOPE;
		if (getServletContext().getAttribute(name) != null) return PageContext.APPLICATION_SCOPE;
		return 0;
	};




    /**
     * @return an enumeration of names (java.lang.String) of all the attributes the specified scope
     */

    public Enumeration getAttributeNamesInScope(int scope) {
		switch (scope) {
			case 1 /* PAGE_SCOPE */ :
				return attr.keys();

			case 2 /* REQUEST_SCOPE */ :
				return ((Pi3HttpServletRequest)req).getAttributeNames();

			case 3 /* SESSION_SCOPE */ :

				/* This is what you can do with Servlet API 2.1 */
/*				if (getSession() != null) {
					String [] names = getSession().getValueNames();
					Vector v = new Vector();
					for (int i = 0; i < Array.getLength(names);) v.add(names[i++]);
					return v.elements();
				} else {
					return new Vector().elements();
				};
*/
				if (getSession() != null) {
					return ((Pi3HttpSession)getSession()).getPi3ValueNames();
				} else {
					return new Vector().elements();
				};

			case 4 /* APPLICATION_SCOPE */ :
				return getServletContext().getAttributeNames();

			default: throw new IllegalArgumentException("Invalid scope.");
		};
	};




    /**
     * @return the current JspWriter stream being used for client response
     */

    public JspWriter getOut() { return out; }




    /**
     * @return the HttpSession for this PageContext or null
     */

    public HttpSession getSession() { 
	    return nSess ? ((HttpServletRequest)req).getSession(true) : null;
	}




    /**
     * @return the Page implementation class instance (Servlet)  associated with this PageContext
     */

    public Object getPage() { return srl; }




    /**
     * @return The ServletRequest for this PageContext
     */

    public ServletRequest getRequest() { return req; } 




    /**
     * @return the ServletResponse for this PageContext
     */

    public ServletResponse getResponse() { return res; }




    /**
     * @return any exception passed to this as an errorpage
     */

    public Exception getException() {
		return (Exception)getRequest().getAttribute("javax.servlet.jsp.jspException");
	};




    /**
     * @return the ServletConfig for this PageContext
     */

    public ServletConfig getServletConfig() { return srl.getServletConfig(); };




    /**
     * @return the ServletContext for this PageContext
     */

    public ServletContext getServletContext() { return srl.getServletConfig().getServletContext(); };




    /**
     * <p>
     * This method is used to re-direct, or "forward" the current ServletRequest and
	 * ServletResponse to another active component in the application.
     * </p>
     * <p>
     * If the <I> relativeUrlPath </I> begins with a "/" then the URL specified
     * is calculated relative to the DOCROOT of the <code> ServletContext </code>
     * for this JSP. If the path does not begin with a "/" then the URL 
     * specified is calculated relative to the URL of the request that was
     * mapped to the calling JSP.
     * </p>
     * <p>
     * It is only valid to call this method from a <code> Thread </code>
     * executing within a <code> _jspService(...) </code> method of a JSP.
     * </p>
     * <p>
     * Once this method has been called successfully, it is illegal for the
     * calling <code> Thread </code> to attempt to modify the <code>
     * ServletResponse </code> object.  Any such attempt to do so, shall result
     * in undefined behavior. Typically, callers immediately return from 
     * <code> _jspService(...) </code> after calling this method.
     * </p>
     *
     * @param relativeUrlPath specifies the relative URL path to the target resource as described above
     *
     * @throws ServletException
     * @throws IOException
     *
     * @throws IllegalArgumentException if target resource URL is unresolvable
     * @throws IllegalStateException if <code> ServletResponse </code> is not in a state where a forward can be performed
     * @throws SecurityException if target resource cannot be accessed by caller
     */

    public void forward(String relativeUrlPath) throws ServletException, IOException {
		RequestDispatcher dsp = getServletContext().getRequestDispatcher(relativeUrlPath);
		if (dsp != null) {
			dsp.forward(req, res);
		} else {
			throw new IllegalArgumentException("Target resource URL is unresolvable.");
		};
	};




    /**
     * <p>
     * Causes the resource specified to be processed as part of the current
     * ServletRequest and ServletResponse being processed by the calling Thread.
     * The output of the target resources processing of the request is written
     * directly to the ServletResponse output stream.
     * </p>
     * <p>
     * The current JspWriter "out" for this JSP is flushed as a side-effect
     * of this call, prior to processing the include.
     * </p>
     * <p>
     * If the <I> relativeUrlPath </I> begins with a "/" then the URL specified
     * is calculated relative to the DOCROOT of the <code> ServletContext </code>
     * for this JSP. If the path does not begin with a "/" then the URL 
     * specified is calculated relative to the URL of the request that was
     * mapped to the calling JSP.
     * </p>
     * <p>
     * It is only valid to call this method from a <code> Thread </code>
     * executing within a <code> _jspService(...) </code> method of a JSP.
     * </p>
     *
     * @param relativeUrlPath specifies the relative URL path to the target resource to be included
     *
     * @throws ServletException
     * @throws IOException
     *
     * @throws IllegalArgumentException if the target resource URL is unresolvable
     * @throws SecurityException if target resource cannot be accessed by caller
     *
     */

    public void include(String relativeUrlPath) throws ServletException, IOException {
		RequestDispatcher dsp = getServletContext().getRequestDispatcher(relativeUrlPath);
		if (dsp != null) {
			dsp.include(req, res);
		} else {
			throw new IllegalArgumentException("Target resource URL is unresolvable.");
		};
	};




    /**
     * <p>
     * This method is intended to process an unhandled "page" level exception
     * by redirecting the exception to either the specified error page for this
     * JSP, or if none was specified, to perform some implementation dependent
     * action.
     * </p>
     * <p>
     * A JSP implementation class shall typically clean up any local state
     * prior to invoking this and will return immediately thereafter. It is
     * illegal to generate any output to the client, or to modify any 
     * ServletResponse state after invoking this call.
     * </p>
     *
     * @param e the exception to be handled
     *
     * @throws ServletException
     * @throws IOException
     *
     * @throws NullPointerException if the exception is null
     * @throws SecurityException if target resource cannot be accessed by caller
     */

    public void handlePageException(Exception e) throws ServletException, IOException {
		
		if (errPage != null) forward(errPage); 
	};
}
