/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2001 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    and Holger Zimmermann (http://pi3web.sourceforge.net/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/TestClnt/StressClient.cpp,v $
 * $Date: 2002/01/19 17:23:32 $
 *
 Description:
	Base class for all C++ handlers, provides a more comfortable interface
	above the raw Pi3 HTTP interface.


\*____________________________________________________________________________*/
//$SourceTop:$

#include <assert.h>
#include <ctype.h>
#include <stdio.h>

#include "StressClient.h"
#include "PIString.h"
#include "DeQuote.h"
#include "PICompat.h"
#include "HTTPDefs.h"
#include "PIStrStr.h"
#include "Pi3API.h"
#include "StrToken.h"

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	StressClient

Description:
	Load and execute an HTTP client handler. The client executes
	the configured test cases in a ordered or a random sequence.
	In collaboration with the Pi3 platform multiprocessing and
	multithreading capabilities high load and stress tests of
	HTTP servers could be performed.

Options:

<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>TestCase
<TD>-
<TD>Method [SP] URI [SP] Protocol
<TD>A simple HTTP request line
<TD>TestCase "GET /index.htm HTTP/1.1"

<TR>
<TD>TestFile
<TD>-
<TD>path to an existing file
<TD>A more complex HTTP request stored in a file.
<TD>TestFile "..\Tests\t100.dat:404"

<TR>
<TD>FileMask
<TD>-
<TD>path to a group of files with wildcards
<TD>A group of HTTP requests, each in a separate file.
<TD>FileMask "..\Tests\*.dat"

<TR>
<TD>Random
<TD>Off
<TD>On/Off
<TD>Sets random or sequential order of requests
<TD>Random On

<TR>
<TD>KeepOpen
<TD>On
<TD>On/Off
<TD>If set, the KeepAlive protocol is used between requests
<TD>KeepOpen Off

<TR>
<TD>KeepOpenCount
<TD>10
<TD>A valid unsigned number
<TD>Number of kept alive HTTP requests
<TD>KeepOpenCount 1000

<TR>
<TD>KeepOpenTimeout
<TD>5
<TD>A valid unsigned number
<TD>Timeout value between kept alive requests in seconds
<TD>KeepOpenTimeout 10

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>TestCase</H5>
This is the simple way to add a new test case to the test client. The HTTP
request line is written directly into the configuration of the StressClient
object. This is only feasable for single line requests, without a body or
additional headers.
<p>The handler will add mandatory headers, if required for the respective
protocol version, e.g. the <i>Host:</i> header in HTTP/1.1.
<p>An expected HTTP status code could be set separated by a colon. If the
server returns another status, a console message is written by the handler.

<H5>TestFile</H5>
This is a more comfortable way to add a new test case to the test client.
The contents of the given text file is sent to the server without additions
by the handler.
<p>An expected HTTP status code could be set separated by a colon. If the
server returns another status, a console message is written by the handler.

<H5>FileMask</H5>
This is the most comfortable way to add a whole group of test cases to the
test client. The file mask will be expanded on startup and all matching files
will be used. The contents of the given text files is sent to the server
as separate requests without additions by the handler.
<p>An expected HTTP status code could be set separated by a colon. If the
server returns another status, a console message is written by the handler.
All requests in a group should expect the same status code in that case.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED, PIAPI_ERROR according to the handler status.

Note:
	The handler could be used with the MultithreadedIOServer and the
	UnixDaemon for high load tests. The underlying TCPIOObject is simply
	switched to active (client) sockets. The address of the server to
	test is configured in the IOObject using the following parameters
	<pre>
	BindHost localhost
	BindPort 80
	</pre>
	<p>If a socket is connected, a new thread is started and this handler
	will be executed. Test cases are performed in a loop until the KeepAlive
	condition becomes false.
	<p>If the tested server does not accept more connections during tests,
	a critical IO error will occur and the Pi3 platform is shut down.
	<p>The test client will be started from the Pi3Client/bin folder -
	<pre>
	./Pi3 ../Conf/StressClient.pi3
	</pre>

Example:
	<PRE>
	&lt;Object&gt;
		Name ClientIOObject
		Class TCPIPIOClass
		Type Active
		BindHost localhost
		BindPort 80
		AcceptTimeout 1000
		RecvTimeout 30
		SendTimeout 30
		PeerIPVariable STRING:RemoteAddr
		PeerHostNameVariable STRING:RemoteHost
		LocalAddressVariable STRING:LocalAddr
		ServerPortVariable RFC822:ServerPort
		Flags OwnDB|SetPeerAddress|SetLocalAddress
	&lt;/Object&gt;

	&lt;Object&gt;
		Name ThreadPoolDispatcher
		Class MultiThreadedIOServerClass
		IOObject ClientIOObject
		LogicObject StressClient
		MaxThreads 5
		ExitAfter 1000
	&lt;/Object&gt;

	&lt;Object&gt;
		Name Main
		Class UNIXDaemonClass
		ServerObject ThreadPoolDispatcher
		PIDFile "../Logs/pid.log"
		RLimitCpu Maximum
		LockFileName "../Logs/lockfile.txt"
		Message "Pi3Web stress test client started"
		NumberOfProcesses 5
	&lt;/Object&gt;

	&lt;Object&gt;
		Name StressClient
		Class StressClientClass
		TestCase "GET /index.htm HTTP/1.1"
		TestCase "GET /pidocs/Features/test4.php3 HTTP/1.1:401"
		TestFile "..\Tests\t10.dat"
		TestFile "..\Tests\t100.dat:404"
		FileMask "..\Tests\*.dat"
		FileMask "..\Tests2\*.dat"
		Random Off
		KeepOpen On
		KeepOpenCount 10
		KeepOpenTimeout 5
	&lt;/Object&gt;
	</PRE>

/*___+++HTMLDOC_END+++___*/
#endif


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
StressClient::StressClient( PIObject *pObject, int iArgc, const char *ppArgv[] ) :
	pObject( pObject ),
	iOK( 1 ),
	iKeepOpenCount( 10 ),
	iProtocol( PR_UNKNOWN ),
	iKeepAlive( 1 ),	
	iAliveTimeout( 50 ),
	pFKRecvTimeout( PIDB_getFastKey( KEY_INT_RECVTIMEOUT, PIDBTYPE_OPAQUE ) ),
	pFKTestfile( PIDB_getFastKey( KEY_INT_TESTFILE, PIDBTYPE_STRING ) ),
	pFKHost( PIDB_getFastKey( KEY_HTTP_HOST, PIDBTYPE_RFC822 ) ),
	iRandom( 0 ),
	iCurrentTest( 0 ),
	iMaxTest( 0 ),
	sUserAgent( "Pi3Web-Client/1.0" ),
	sVirtualHost( 0 ),
	logfile( PIPLATFORM_FD_INVALID ),
	sLogfile( 0 ),
	sOpenMode( "a" ),
	pTestcaseDB( PIDB_new( 0, "StressTests") )
{
    iOK = PIObject_readParameters( pObject, iArgc, ppArgv, ParameterFn,	this );

	HTTPCore_initClient();

	// Adjust testcase counters
	if (iCurrentTest)
		{
		iMaxTest = iCurrentTest - 1;
		iCurrentTest = 0;
		};

}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
StressClient::~StressClient()
{
	HTTPCore_cleanupClient();
	PIDB_delete( pTestcaseDB );
	pTestcaseDB = 0;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Verify the HTTP request line and divide it into components. Returns
	INT_CONTINUE on success or an HTTP error status code on error.

	Assumes pLine is NULL terminated at pLine[iLineLen].

	On success instantiates all values to at least an empty string with the
	exception of the proxy values which will all be instantiated if
	ppProxyProtocol is instantiated.

\*____________________________________________________________________________*/
static int SplitRequestLine(
	char *pLine, 
	int iLineLen,
	const char **ppMethod,
	const char **ppProxyProtocol,
	const char **ppProxyHost,
	int *piProxyHostLen,
	const char **ppProxyPort,
	int *piProxyPortLen,
	const char **ppPath,
	const char **ppQueryString,
	const char **ppProtocol
	)
{
	assert( pLine );
	assert( iLineLen );
	assert( ppMethod );
	assert( ppProxyProtocol );
	assert( ppProxyHost );
	assert( piProxyHostLen );
	assert( ppProxyPort );
	assert( piProxyPortLen );
	assert( ppPath );
	assert( ppProtocol );

	*ppProtocol = "";	/* --- this is done because protcol is optional, i.e.
						HTTP/0.9 doesn't send one
						--- */
	/* ---
	method
	--- */
	int i = 0;
	for( ; i<iLineLen && !(isspace(pLine[i])); i++);
	*ppMethod = pLine;
	if ( i==iLineLen ) { return ST_BADREQUEST; };
	pLine[i]='\0'; i++;
	for(; i<iLineLen && (isspace(pLine[i])); i++);
	pLine = &( pLine[i] ); iLineLen -= i; i=0; 

	/* ---
	pLine now points at the component following the method
	--- */

	/* --- URL path must exist --- */
	if ( i==iLineLen )
		{ return ST_BADREQUEST; };

	/* ---
	proxy host and port
	--- */
	if ( (*pLine!='/') && (*pLine!='*'))
		{
		/* --- assume proxy host and port specification --- */
		for(;
			(i+2)<iLineLen &&
			strncmp(&(pLine[i]), "://", 3) && !(isspace(pLine[i])); i++)
			{ pLine[i] = tolower( pLine[i] ); };
		if ( !strncmp( &(pLine[i]), "://", 3 ) )
			{
			*ppProxyProtocol = pLine; pLine[i] = '\0';
			i += 3;
			pLine = &( pLine[i] ); iLineLen -= i; i=0;

			/* --- pLine now points at proxy host --- */
			*ppProxyHost = pLine;
			for(;	i<iLineLen &&
					pLine[i]!=':' &&
					pLine[i]!='/' &&
					!(isspace(pLine[i])); i++);
			*piProxyHostLen = i;
			if ( pLine[i]==':' ) 
				{
				i++;
				pLine = &( pLine[i] ); iLineLen -= i; i=0;

				/* --- pLine now points at proxy port --- */
				*ppProxyPort = pLine;
				for(; i<iLineLen && pLine[i]!='/' && !(isspace(pLine[i])); i++);
				*piProxyPortLen = i;
				}
			else
				{
				*ppProxyPort = "";
				*piProxyPortLen = 0;
				};
			pLine = &( pLine[i] ); iLineLen -= i; i=0;
			}
		else
			{
			/* ---
			The URL path didn't start with '/' but it wasn't a proxy request

			NOTE: Maybe this should raise a gateway error instead of the
			generic bad request error.

			--- */
			return ST_BADREQUEST;
			};
		};

	/* ---	
	pLine now points at first character of URL path
	--- */

	/* ---
	path and query string.
	--- */
	for(; i<iLineLen && pLine[i]!='?' && !(isspace(pLine[i])); i++);
	*ppPath = pLine;

	if ( pLine[i]=='?' )
		{ 
		pLine[i++] = '\0';
		*ppQueryString = &(pLine[i]);
		for(; i<iLineLen && !(isspace(pLine[i])); i++);
		}
	else
		{ *ppQueryString = ""; };

	if ( i==iLineLen ) { return INT_CONTINUE; };
	pLine[i++]='\0';
	for(; i<iLineLen && (isspace(pLine[i])); i++);
	pLine = &( pLine[i] ); iLineLen -= i; i=0; 

	/* ---
	pLine now points at the component following the URL path and
	query string.
	--- */

	/* ---
	protocol is last component on the line, skip components until the last
	--- */
	for(;;)
		{
		for(; i<iLineLen && !(isspace(pLine[i])); i++);
		*ppProtocol = pLine;
		for(; i<iLineLen && (isspace(pLine[i])); i++);
		if ( i==iLineLen ) { return INT_CONTINUE; }; /* this should be true */

		/* --- I guess not --- */
		pLine = &( pLine[i] ); iLineLen -= i; i=0;
		};

	/* -- OK --- */
	return INT_CONTINUE;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int StressClient::ReadRequestLine( PIDB *pDB, const char *pLine )
	{
	const char *pMethod = 0;
	const char *pProxyProtocol = 0;
	const char *pProxyHost = 0;
	int iProxyHostLen = 0;
	const char *pProxyPort = 0;
	int iProxyPortLen = 0;
	const char *pURI = 0;
	const char *pQueryString = 0;
	const char *pProtocol = 0;

	int iStatus = SplitRequestLine(
		(char *)pLine, 
		strlen(pLine),
		&pMethod,
		&pProxyProtocol,
		&pProxyHost,
		&iProxyHostLen,
		&pProxyPort,
		&iProxyPortLen,
		&pURI,
		&pQueryString,
		&pProtocol
		);
		
	if ( iStatus ) { return iStatus; };

	if ( !pMethod || !pURI || !pQueryString || !pProtocol )
		{ return ST_BADREQUEST; };

	/* --- store these headers --- */
	PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_METHOD, (void *)pMethod, 0 );
	PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_URI, (void *)pURI, 0 );
	PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_QUERYSTRING, (void *)pQueryString, 0 );
	PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_PROTOCOL, (void *)pProtocol, 0 );

	/* --- Is this a proxy request? --- */
	if ( pProxyProtocol )
		{
		PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_PROXYPROTOCOL,
			(void *)pProxyProtocol, 0 );
		PIString sProxyHost( pProxyHost, iProxyHostLen );
		PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_PROXYHOST,
			(void *)(const char *)sProxyHost, 0 );
		PIString sProxyPort( pProxyPort, iProxyPortLen );
		PIDB_add( pDB, PIDBTYPE_STRING, KEY_HTTP_PROXYPORT,
			(void *)(const char *)sProxyPort, 0 );
		};

	/* ---
	Convert method and status into internal numeric form	
	--- */
	int iMethod = MD_UNKNOWN;
	int iProtocol = PR_UNKNOWN;
	if ( !PIUtil_stricmp( pMethod, MD_NAME_GET ) )
		{ iMethod = MD_GET; }
	else if ( !PIUtil_stricmp( pMethod, MD_NAME_POST ) )
		{ iMethod = MD_POST; }
	else if ( !PIUtil_stricmp( pMethod, MD_NAME_HEAD ) )
		{ iMethod = MD_HEAD; }
	else if ( !PIUtil_stricmp( pMethod, MD_NAME_PUT ) )
		{ iMethod = MD_PUT; }
	else if ( !PIUtil_stricmp( pMethod, MD_NAME_DELETE ) )
		{ iMethod = MD_DELETE; }
	else if ( !PIUtil_stricmp( pMethod, MD_NAME_OPTIONS ) )
		{ iMethod = MD_OPTIONS; }
	else if ( !PIUtil_stricmp( pMethod, MD_NAME_TRACE ) )
		{ iMethod = MD_TRACE; };

	if ( *pProtocol )
		{
		/* --- any protocol is not HTTP/0.9 --- */
		if ( !PIUtil_stricmp( pProtocol, PR_NAME_HTTP10 ) )
			{ iProtocol = PR_HTTP10; }
		else if ( !PIUtil_stricmp( pProtocol, PR_NAME_HTTP11 ) )
			{ iProtocol = PR_HTTP11; };
		}
	else
		{ iProtocol = PR_HTTP09; };

	/* --- Do we support this method and protocol ? --- */
	if ( iMethod == MD_UNKNOWN ) { return ST_NOTIMPLEMENTED; };
	if ( iProtocol == PR_UNKNOWN ) { iProtocol = PR_HTTP10; };

	/* --- Do we support the method for version < HTTP/1.1 ? --- */
	if ((iProtocol != PR_HTTP11) && (iMethod > MD_HEAD))
		{ return ST_BADREQUEST; }; // maybe better ST_NOTIMPLEMENTED?
	
	/* --- set the numeric method and protocol --- */
	PIDB_add( pDB, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD, (void *)iMethod, 0 );
	PIDB_add( pDB, PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL, (void *)iProtocol, 0 );

	return iStatus;
	}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int StressClient::ReadFileMask( const char *pValue, int iStatus )
{

	PIFInfo *pInfo = PIFInfo_new( pValue );

	PIString sFilePath = PIFInfo_getPathRoot( pInfo );
	PIString sFileMask = PIFInfo_getName( pInfo );

	if ( sFileMask && sFilePath )
		{
		PIFInfo *pDir = HTTPCore_getCachedFile( (const char *)sFilePath );
		if ( pDir && PIFInfo_isDirectory( pDir ) )
			{

			PIFInfo *pFile = PIFInfo_getFirstFileInDirectory( pDir );
			enum { BUFSIZE=15 };
			char szBuf[BUFSIZE+1];
		
			while( pFile )
				{
	
				/* ---
				Check file is allowed, if mask pattern matches
				then include this file
				--- */
				if ( PIFInfo_isReadable( pFile ) )
					{

					/* ---
					Get filename
					--- */
					PIString sFileName( PIFInfo_getName( pFile ) );

					if ( HTTPUtil_regexMatch( sFileMask, sFileMask.Len(),
						sFileName, sFileName.Len() ))
						{
						sprintf( szBuf, "%lu", iCurrentTest++ );
						PIString sTestFile = sFilePath;
						sTestFile.Concatenate( PIPlatform_getDirectorySeparator());
						sTestFile.Concatenate( sFileName );
						PIDB *pDB = PIDB_new( pTestcaseDB, szBuf );
						PIDB_add( pDB, PIDBTYPE_STRING, KEY_INT_TESTFILE,
							(void *)(const char *)sTestFile, 0 );
						PIDB_add( pDB, PIDBTYPE_OPAQUE, KEY_INT_STATUS,
							(void *)iStatus, 0 );
						};
					};

				if ( !PIFInfo_getNextFileInDirectory( pFile ) )
					{ PIFInfo_delete( pFile ); pFile = 0; };

				}; //while
			HTTPCore_releaseCachedFile( pDir );
			};
		};
	return PIAPI_COMPLETED;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int StressClient::Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
	{
	assert( pVariable && pValue );
	PIOStrStream os;
	os << pWhere << "StressClient: ";

	if ( !PIUtil_stricmp( KEY_CONF_USERAGENT, pVariable ) )
		{
		sUserAgent = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_PROTOCOL, pVariable ) )
		{
		iProtocol = (!PIUtil_stricmp((const char *)DeQuote( pValue ),
			"HTTP/1.1")) ? PR_HTTP11 : PR_HTTP10;
		}
	else if ( !PIUtil_stricmp( KEY_CONF_VIRTUALHOST, pVariable ) )
		{
		sVirtualHost = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_KEEPOPEN, pVariable ) )
		{
		iKeepAlive = !PIUtil_stricmp((const char *)DeQuote( pValue ), "On");
		}
	else if ( !PIUtil_stricmp( KEY_CONF_KEEPOPENTIMEOUT, pVariable ) )
		{
		iAliveTimeout = atoi((const char *)DeQuote( pValue ));
		}
	else if ( !PIUtil_stricmp( KEY_CONF_KEEPOPENCOUNT, pVariable ) )
		{
		iKeepOpenCount = atoi((const char *)DeQuote( pValue ));
		}
	else if ( !PIUtil_stricmp( KEY_CONF_RANDOM, pVariable ) )
		{
		iRandom = !PIUtil_stricmp((const char *)DeQuote( pValue ), "On");
		}
	else if ( !PIUtil_stricmp( KEY_CONF_FILEMASK, pVariable ) )
		{
		StringTokenizer tTokens( (const char *)DeQuote( pValue ),":" );
		int iStatus = tTokens.NumTokens() > 1 ? atoi( tTokens.GetToken( 1 ) ) : 200;
		ReadFileMask( (const char *)tTokens.GetToken( 0 ), iStatus );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_TESTFILE, pVariable ) )
		{
		enum { BUFSIZE=15 };
		char szBuf[BUFSIZE+1];
		sprintf( szBuf, "%lu", iCurrentTest++ );
		StringTokenizer tTokens( (const char *)DeQuote( pValue ), ":" );
		PIDB *pDB = PIDB_new( pTestcaseDB, szBuf );
		PIDB_add( pDB, PIDBTYPE_STRING, KEY_INT_TESTFILE,
			(void *)(const char *)tTokens.GetToken( 0 ), 0 );
		int iStatus = tTokens.NumTokens() > 1 ? atoi( tTokens.GetToken( 1 ) ) : 200;
		PIDB_add( pDB, PIDBTYPE_OPAQUE, KEY_INT_STATUS, (void *)iStatus, 0 );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_TESTCASE, pVariable ) )
		{
		enum { BUFSIZE=15 };
		char szBuf[BUFSIZE+1];
		sprintf( szBuf, "%lu", iCurrentTest++ );
		PIDB *pDB = PIDB_new( pTestcaseDB, szBuf );
		StringTokenizer tTokens( (const char *)DeQuote( pValue ), ":" );

		if ( !ReadRequestLine( pDB,	(const char *)tTokens.GetToken( 0 ) ) == INT_CONTINUE )
			{
			// ErrorMsg!
			iOK = 0;
			}
		else
			{
			int iStatus = tTokens.NumTokens() > 1 ? atoi( tTokens.GetToken( 1 ) ) : 200;
			PIDB_add( pDB, PIDBTYPE_OPAQUE, KEY_INT_STATUS,
				(void *)iStatus, 0 );
			};
		}
	else if ( !PIUtil_stricmp( KEY_CONF_LOGFILE, pVariable ) )
		{
		sLogfile = (const char *)DeQuote( pValue );
		}
	else if ( !PIUtil_stricmp( KEY_CONF_OPENMODE, pVariable ) )
		{
		sOpenMode = (const char *)DeQuote( pValue );
		}
	else
		{
		os << "Unknown directive '" << pVariable <<
			"'" << ends;
		CONFIG_ERR( Object(), os.str() );
		return 0;
		};

	return 1;
	};

int hex2int(char *s)
{
	int iRes = 0;
	for (unsigned int i=0;i<strlen(s);i++)
		{
		s[i] = toupper(s[i]);
		iRes <<= 4;
		iRes += (s[i]<'A') ? s[i]-48 : s[i]-55;
		}
	return iRes;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int StressClient::Execute( int iArgc, const char *ppArgv[] )
{
	PI_MARK;

	if ( iArgc<1 ) { return PIAPI_ERROR; };
	PIObject *pIO = (PIObject *)*ppArgv;
	if ( !pIO )	{ return PIAPI_ERROR; };

	/* ---
	Get the DB from the IO object to monitor the KeepOpen flag
	--- */
	PIDB *pC = PIObject_getDB( pIO );
	int iRet;
	int iTimeout;

	PIIOBuffer *pBuffer = PIIOBuffer_new( pIO );
	if ( !pBuffer ) return PIAPI_ERROR;

	if ( iKeepAlive ) {
		assert( pFKRecvTimeout );
		iTimeout = (int)PIDB_lookup( pC, PIDBTYPE_OPAQUE,
			pFKRecvTimeout, PIDBFLAG_FASTKEY );

		PIDB_replace( pC, PIDBTYPE_OPAQUE, pFKRecvTimeout,
			(void *)iAliveTimeout, PIDBFLAG_FASTKEY );
	};

	if (iRandom) srand( (unsigned)time( NULL ) );

    int iDoKeepOpen = iKeepAlive;
	int iDoKeepOpenCount = iKeepOpenCount;
	int iServerProtocol = PR_HTTP11;
	const char *pKeepOpen = 0;
    PIDB *pR = PIDB_new( 0, "R" );

	for(;;)
		{
		/* ---
		Disallow more than the specified number of connection reuses
		--- */
		if ( !iDoKeepOpenCount )
			iDoKeepOpen = 0;
		else
			iDoKeepOpenCount--;

		/* ---
		Get the testcase from DB
		--- */
		if ( !pTestcaseDB ) break;
		enum { BUFSIZE=16 };
		char szBuf[BUFSIZE+1];
		if ( iRandom )
			{
			iCurrentTest = rand() % (iMaxTest + 1);
			}
		else
			{
			if (iCurrentTest > iMaxTest ) { iCurrentTest = 0; };
			}
		sprintf( szBuf, "%lu", iCurrentTest );
		PIDB *pQ = (PIDB *)PIDB_lookup( pTestcaseDB, PIDBTYPE_TREE, szBuf, 0 );
		if ( !pQ ) return PIAPI_ERROR;

		/* --- get ready for new request --- */
		const char *pPath = (const char *)PIDB_lookup( pQ, PIDBTYPE_STRING,
			pFKTestfile, PIDBFLAG_FASTKEY );

		iRet = pPath ? SendRequestFile( pBuffer, pPath )
		             : SendRequestDB( pBuffer, pC, pQ, iDoKeepOpen );

	    /* --- send the whole thing on its way --- */
		PIIOBuffer_flush( pBuffer );
		iRet = PIIOBuffer_pollBeforeRead( pBuffer );
		if (iRet < 0)
			{
			cout << "PIIOBuffer_pollBeforeRead failed" <<
				endl << "Testcase: " << iCurrentTest << endl;
			goto exit_handle;
			}

	    /*
		** Read the response
	    */
		iRet = HTTPCore_readHeaders( pBuffer, pR, RH_RESPONSE );
		if ( iRet )
			{
			cout << "HTTPCore_readHeaders failed: " << iRet <<
				endl << "Testcase: " << iCurrentTest << endl;
	        goto exit_handle;
		    };
    
		/* ---
	    Check out status line from remote server
		--- */
		int iStatusCode = (int)PIDB_lookup( pR, PIDBTYPE_OPAQUE, KEY_HTTP_STATUSCODE, 0 );
		if ( iStatusCode !=	(int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_INT_STATUS, 0 ))
			{
			cout << "Server returned status " << iStatusCode <<	endl
				 << "expected status "
				 << (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_INT_STATUS, 0 )
				 << " for Testcase " << iCurrentTest << endl;
			};
	    
	    /*
		** Read the rest of the response
	    */

		/* --- get length to read --- */
		unsigned long iRead = 0;
		unsigned long iToRead;
		int iBufRead;
		const char *pData;
		const char *pTE = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			KEY_HTTP_TRANSFERENCODING, 0 );
		if ( pTE && !PIUtil_stricmp( pTE, KEY_HTTP_CHUNKED ))
			{
			pData = (const char *)PIUtil_malloc(0x4000);

			/* --- loop until no more bytes left --- */
			do
				{
				/* --- start new chunk --- */				
				iRead = 0;

				/* --- get chunk header with data size --- */				
				PIIOBuffer_getLine( pBuffer, szBuf, BUFSIZE );
				iToRead = hex2int(szBuf);
				if ( !iToRead )
					{
					PIIOBuffer_getLine( pBuffer, szBuf, BUFSIZE );
					break;
					};

				if ( iToRead > 0x4000 )
					{
					cout << "Received chunksize to large for internal buffer" << endl;
					iRet = 400;
					break;
					};

				/* --- read the whole chunk --- */				
				do
					{

					/* --- no more data or client timeout --- */
					if ( PIIOBuffer_pollBeforeRead( pBuffer ) < 0 ) break;

					/* --- read the data --- */
					iBufRead = PIIOBuffer_readToBuffer( pBuffer, (char *)pData,
						iToRead - iRead );
					if ( iRead == -1)
						{
						cout << "Error reading chunk into internal buffer" << endl;
						iRet = 400;
						break;
						};
					iRead += iBufRead;
					}
				while ( iToRead - iRead );

				PIIOBuffer_getLine( pBuffer, szBuf, BUFSIZE );
				}
			while ( iToRead );
			PIUtil_free( (void *)pData );
			}
		else
			{
			pData = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
				KEY_HTTP_CONTENTLENGTH, 0 );
			iToRead = pData ? strtol(pData, NULL, 10) : 0;

			/* --- loop until no more bytes left --- */
			do
				{
				if ( !iToRead ) break;

				/* --- no more data or client timeout --- */
				if ( PIIOBuffer_pollBeforeRead( pBuffer ) < 0 ) break;

				/* --- read the data --- */
				pData = PIIOBuffer_read( pBuffer, &iBufRead );
				iRead += iBufRead;

				/* --- this is an error --- */
				if ( !iBufRead ) break;
				}
			while ( iRead < iToRead );

			if (iRead < iToRead)
				{
				cout << "Could not read server response body" <<
					endl << "Testcase: " << iCurrentTest << endl;
				iRet = 400;
				goto exit_handle;
				};

			if (iRead > iToRead)
				{
				cout << "Server sent to much data in response" <<
					endl << "Testcase: " << iCurrentTest << endl;
				iRet = 400;
				goto exit_handle;
				};
			};

		pKeepOpen = (const char *)PIDB_lookup( pR, PIDBTYPE_RFC822,
			KEY_HTTP_CONNECTION, 0 );
		iServerProtocol = (int)PIDB_lookup( pR, PIDBTYPE_OPAQUE,
			KEY_HTTP_PROTOCOL, 0 );
		iDoKeepOpen = (( pKeepOpen && !PIUtil_stricmp( pKeepOpen, "Keep-Alive" ))
			|| ( !pKeepOpen && iServerProtocol == PR_HTTP11 )) ? 1: 0;

		PIDBIterator *pIter = PIDB_getIterator( pR, PIDBTYPE_RFC822, 0, 0 );
		if ( pIter ) {
			while ( PIDBIterator_atValidElement( pIter ) ) {
				PIDBIterator_removeCurrent( pIter );
			};
			PIDBIterator_delete( pIter );
		};

		pIter = PIDB_getIterator( pR, PIDBTYPE_OPAQUE, 0, 0 );
		if ( pIter ) {
			while ( PIDBIterator_atValidElement( pIter ) ) {
				PIDBIterator_removeCurrent( pIter );
			};
			PIDBIterator_delete( pIter );
		};

		iCurrentTest++;
		if ( iRet != PIAPI_COMPLETED || !iDoKeepOpen ) { break; };
		}; // for(;;)

		if ( iKeepAlive ) {
			PIDB_replace( pC, PIDBTYPE_OPAQUE, pFKRecvTimeout,
				(void *)iTimeout, PIDBFLAG_FASTKEY );
		};

exit_handle:
	PIIOBuffer_delete( pBuffer );
	PIDB_delete( pR );
	PI_MARK;
	return iRet;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static const char *pProtocolMap[] = 
{
    PR_NAME_UNKNOWN,
    PR_NAME_HTTP09,
    PR_NAME_HTTP10,
    PR_NAME_HTTP11,
    0
};
int StressClient::SendRequestDB(PIIOBuffer *pBuffer, PIDB *pC, PIDB *pQ, int iKeepOpen )
	{
	int iRet = PIAPI_COMPLETED;

    /*
    ** Add some headers
    */
	PIDB_replace( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_CONNECTION, (void *)iKeepOpen, 0 );
	PIDB_replace( pQ, PIDBTYPE_RFC822, KEY_HTTP_CONTENTLENGTH, (void *)"0", 0 );

	if (iProtocol != PR_UNKNOWN)
		{
		PIDB_replace( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL,
			(void *)iProtocol, 0 );
		PIDB_replace( pQ, PIDBTYPE_STRING, KEY_HTTP_PROTOCOL,
			(void *)pProtocolMap[iProtocol], 0);
		};	

	if ( sUserAgent.Len() )
		{
		PIDB_replace( pQ, PIDBTYPE_RFC822, "User-Agent",
			(void *)(const char *)sUserAgent, 0 );
		};

	if ( sVirtualHost.Len() )
		{
		PIDB_replace( pQ, PIDBTYPE_RFC822, pFKHost,
			(void *)(const char *)sVirtualHost, PIDBFLAG_FASTKEY );
		};

    /* ---
	Make sure the 'host' header is sent
	--- */
	if (!PIDB_lookup( pQ, PIDBTYPE_RFC822, pFKHost, PIDBFLAG_FASTKEY ) )
		{
	    const char *pPort = (const char *)PIDB_lookup( pC, PIDBTYPE_RFC822,
			KEY_INT_SERVERPORT, 0);
		if (!pPort) pPort = "80";

		const char *pHost = (const char *)PIDB_lookup( pC, PIDBTYPE_STRING,
			KEY_INT_REMOTEHOST, 0);
		if (!pHost) pHost = "localhost";

		PIOStrStream os;
		os << pHost;
		if ( atoi(pPort) != HTTP_DEFAULT_PORT )
	        {
		    os << ":" << pPort;
			};
	    os << ends;

		PIDB_replace( pQ, PIDBTYPE_RFC822, pFKHost, (void *)os.str(), PIDBFLAG_FASTKEY );
		};

    /*
    ** Make the request
    */
    iRet = HTTPUtil_doHTTPRequest( pBuffer, pQ );
    if ( iRet != PIAPI_COMPLETED )
        {
		cout << "doHTTPRequest failed" << endl;
        };

	return iRet;
	}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int StressClient::SendRequestFile(PIIOBuffer *pBuffer, const char *pPath )
	{
    int iRet = PIAPI_COMPLETED;
	
	/* ---
	Get file object
	--- */
	PIFInfo *pFInfo = HTTPCore_getCachedFile( pPath );
	if ( !pFInfo ) return PIAPI_ERROR;

	/* ---	
	This file must exist and be a regular file 
	--- */
	if ( !PIFInfo_exists( pFInfo ) || !PIFInfo_isRegular( pFInfo ) )
		{
		HTTPCore_releaseCachedFile( pFInfo );	
		cout << "PIFInfo_exists || PIFInfo_isRegular failed" << endl;
		return PIAPI_ERROR;
		};

	/* ---	
	Send the file
	--- */
	iRet = HTTPUtil_sendFile( pBuffer, pFInfo, PIIOBUF_NONE, 0 );
	if (iRet != PIAPI_COMPLETED)
	    {
		cout << "HTTPUtil_sendFile returned: " << iRet << endl;
	    }

	HTTPCore_releaseCachedFile( pFInfo );
	return iRet;
	}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StressClient_execute( PIObject *pObj, int iArgc,
		const char *ppArgv[] )
{
	if ( !pObj ) return PIAPI_ERROR;
	return ((StressClient *)PIObject_getUserData(pObj))->Execute( iArgc, ppArgv );
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StressClient_onClassLoad( PIClass_LoadAction eLoad, void * )
{
	switch( eLoad )
		{
		case STARTUP:
		default:;
		};
	return PIAPI_COMPLETED;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StressClient_constructor( PIObject *pObj, int iArgc, const char *ppArgv[] )
{
	StressClient *pStressClient = new StressClient( pObj, iArgc, ppArgv );

	if (!( pStressClient && pStressClient->IsOK() ))
		{
		delete pStressClient;
		return PIAPI_ERROR;
		};

	if ( PIObject_setUserData( pObj, pStressClient ) )
		{
		delete pStressClient;
		CONFIG_ERR( pObj, "StressClient_constructor: PIObject_setUserData() failed" );
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StressClient_copyConstructor( PIObject *pObj, int, const char *[] )
{
	/* --- forbid copy constructor --- */
	return PIAPI_ERROR;
}


/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int StressClient_destructor( PIObject *pObj, int, const char *[] )
{
	delete (StressClient *)PIObject_getUserData( pObj );
	return PIAPI_COMPLETED;
}

#if 0
/*___+++CNF_BEGIN+++___*/

	<Class>
		Name StressClientClass
		Type LogicExtension
		Library StressClient
		OnClassLoad StressClient_onClassLoad
		Constructor StressClient_constructor
		CopyConstructor StressClient_copyConstructor
		Destructor StressClient_destructor
		Execute StressClient_execute
	</Class>

	<Object>
		Name StressClient
		Class StressClientClass
		TestCase "GET /?param1=value1 HTTP/1.1"
		TestCase "GET /icons/Pi3Tile.gif HTTP/1.1"
		TestCase "GET /icons/Pi3Web_earth1.gif HTTP/1.1"
		TestCase "GET /index.htm HTTP/1.1"
		TestCase "GET /samples.htm HTTP/1.0"
#		TestCase "GET /pidocs/Features/test1.php3 HTTP/1.0"
#		TestCase "GET /pidocs/Features/test2.php3 HTTP/1.1"
#		TestCase "GET /pidocs/Features/test3.php3 HTTP/1.1"
#		TestCase "GET /pidocs/Features/test4.php3 HTTP/1.1:401"
#		TestFile "..\Tests\t1.dat"
#		TestFile "..\Tests\t10.dat"
#		TestFile "..\Tests\t100.dat:404"
#		TestFile "..\Tests\t11.dat:401"
#		TestFile "..\Tests\t12.dat"
#		TestFile "..\Tests\t13.dat:301"
#		TestFile "..\Tests\t14.dat"
#		TestFile "..\Tests\t15.dat"
#		TestFile "..\Tests\t16.dat"
#		TestFile "..\Tests\t17.dat"
#		TestFile "..\Tests\t18.dat"
#		TestFile "..\Tests\t19.dat"
#		TestFile "..\Tests\t2.dat"
#		TestFile "..\Tests\t3.dat:301"
#		TestFile "..\Tests\t31.dat:500"
#		TestFile "..\Tests\t4.dat"
#		TestFile "..\Tests\t41.dat:301"
#		TestFile "..\Tests\t42.dat:301"
#		TestFile "..\Tests\t43.dat:301"
#		TestFile "..\Tests\t44.dat:301"
#		TestFile "..\Tests\t45.dat:301"
#		TestFile "..\Tests\t5.dat"
#		TestFile "..\Tests\t50.dat"
#		TestFile "..\Tests\t6.dat"
#		TestFile "..\Tests\t61.dat"
#		TestFile "..\Tests\t7.dat"
#		TestFile "..\Tests\t71.dat"
#		TestFile "..\Tests\t8.dat"
#		TestFile "..\Tests\t81.dat"
#		TestFile "..\Tests\t9.dat"
#		FileMask "..\Tests\*.dat"
#		FileMask "..\Tests2\*.dat"
		Random Off
		KeepOpen On
		KeepOpenCount 10
		KeepOpenTimeout 5
	</Object>

/*___+++CNF_END+++___*/
#endif
